from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

from .aurora import nflx_aurora_capacity_model
from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import FixedInterval
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import Platform
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.models import CapacityModel


class NflxPostgresCapacityModel(CapacityModel):
    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:
        if desires.service_tier == 0:
            return None

        plan = None
        if Platform.aurora_postgres in instance.platforms:
            plan = nflx_aurora_capacity_model.capacity_plan(
                instance=instance,
                drive=drive,
                context=context,
                desires=desires,
                extra_model_arguments=extra_model_arguments,
            )
        return plan

    @staticmethod
    def description():
        return "Netflix Postgres Model"

    @staticmethod
    def allowed_platforms() -> Tuple[Platform, ...]:
        return Platform.aurora_postgres, Platform.amd64

    @staticmethod
    def default_desires(user_desires, extra_model_arguments):
        return CapacityDesires(
            query_pattern=QueryPattern(
                access_pattern=AccessPattern.latency,
                access_consistency=GlobalConsistency(
                    same_region=Consistency(
                        target_consistency=AccessConsistency.serializable_stale,
                    ),
                    cross_region=Consistency(
                        target_consistency=AccessConsistency.never,
                    ),
                ),
                # can't really make latency/throughput trade-offs with RDS
                estimated_mean_read_size_bytes=Interval(
                    low=128, mid=1024, high=65536, confidence=0.90
                ),
                estimated_mean_write_size_bytes=Interval(
                    low=64, mid=512, high=2048, confidence=0.90
                ),
                estimated_mean_read_latency_ms=Interval(
                    low=1, mid=4, high=100, confidence=0.90
                ),
                estimated_mean_write_latency_ms=Interval(
                    low=1, mid=6, high=200, confidence=0.90
                ),
                read_latency_slo_ms=FixedInterval(
                    minimum_value=1,
                    maximum_value=100,
                    low=1,
                    mid=10,
                    high=20,
                    confidence=0.98,
                ),
                write_latency_slo_ms=FixedInterval(
                    minimum_value=1,
                    maximum_value=100,
                    low=1,
                    mid=10,
                    high=20,
                    confidence=0.98,
                ),
            ),
            # Assume that the working set is between 20% by default
            data_shape=DataShape(
                estimated_working_set_percent=Interval(
                    low=0.05, mid=0.10, high=0.20, confidence=0.8
                )
            ),
        )


nflx_postgres_capacity_model = NflxPostgresCapacityModel()



# -*- coding: utf-8 -*-
import functools
import logging
import math
from hashlib import blake2b
from typing import Any
from typing import Callable
from typing import cast
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple

import numpy as np

from service_capacity_modeling.hardware import HardwareShapes
from service_capacity_modeling.hardware import shapes
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import CapacityRegretParameters
from service_capacity_modeling.interface import CapacityRequirement
from service_capacity_modeling.interface import certain_float
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import Hardware
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import interval
from service_capacity_modeling.interface import Lifecycle
from service_capacity_modeling.interface import PlanExplanation
from service_capacity_modeling.interface import Platform
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.interface import Requirements
from service_capacity_modeling.interface import UncertainCapacityPlan
from service_capacity_modeling.models import CapacityModel
from service_capacity_modeling.models.common import merge_plan
from service_capacity_modeling.models.org import netflix
from service_capacity_modeling.models.utils import reduce_by_family
from service_capacity_modeling.stats import dist_for_interval
from service_capacity_modeling.stats import interval_percentile

logger = logging.getLogger(__name__)


def simulate_interval(
    interval: Interval, name: str
) -> Callable[[int], Sequence[Interval]]:
    if interval.can_simulate:
        # We need to convert the name of the field to a positive 32 bit seed
        # and we don't really need collision prevention so just take a 24 bit
        # digest and enforce it is not signed
        seed = int.from_bytes(
            blake2b(name.encode(), digest_size=3).digest(),
            byteorder="big",
            signed=False,
        )

        def sim_uncertain(count: int) -> Sequence[Interval]:
            return [
                certain_float(s)
                for s in dist_for_interval(interval, seed=seed).rvs(count)
            ]

        return sim_uncertain

    else:

        def sim_certain(count: int) -> Sequence[Interval]:
            return [interval] * count

        return sim_certain


# Take uncertain inputs and simulate a desired number of certain inputs
# e.g. read_per_second[100, 1000] -> [rps[107, 107], rps[756, 756], ...]
# num_sims concrete values are generated.
def model_desires(
    desires: CapacityDesires, num_sims: int = 1000
) -> Generator[CapacityDesires, None, None]:
    query_pattern = desires.query_pattern.model_copy(deep=True)
    data_shape = desires.data_shape.model_copy(deep=True)

    query_pattern_simulation = {}
    for field in sorted(query_pattern.model_fields):
        d = getattr(query_pattern, field)
        if isinstance(d, Interval):
            query_pattern_simulation[field] = simulate_interval(d, field)(num_sims)
        else:
            query_pattern_simulation[field] = [d] * num_sims

    data_shape_simulation = {}
    for field in sorted(data_shape.model_fields):
        d = getattr(data_shape, field)
        if isinstance(d, Interval):
            data_shape_simulation[field] = simulate_interval(d, field)(num_sims)
        else:
            data_shape_simulation[field] = [d] * num_sims

    for sim in range(num_sims):
        query_pattern = QueryPattern(
            **{
                f: query_pattern_simulation[f][sim]
                for f in sorted(query_pattern.model_fields)
            }
        )
        data_shape = DataShape(
            **{
                f: data_shape_simulation[f][sim]
                for f in sorted(data_shape.model_fields)
            }
        )

        d = desires.model_copy(update={"query_pattern": None, "data_shape": None})
        d.query_pattern = query_pattern
        d.data_shape = data_shape
        yield d


def model_desires_percentiles(
    desires: CapacityDesires,
    percentiles: Sequence[int] = (5, 25, 50, 75, 95),
) -> Tuple[Sequence[CapacityDesires], CapacityDesires]:
    query_pattern = desires.query_pattern.model_copy(deep=True)
    data_shape = desires.data_shape.model_copy(deep=True)

    query_pattern_simulation = {}
    query_pattern_means = {}
    for field in sorted(query_pattern.model_fields):
        d = getattr(query_pattern, field)
        if isinstance(d, Interval):
            query_pattern_simulation[field] = interval_percentile(d, percentiles)
            if d.can_simulate:
                query_pattern_means[field] = certain_float(d.mid)
            else:
                query_pattern_means[field] = d
        else:
            query_pattern_simulation[field] = [d] * len(percentiles)
            query_pattern_means[field] = d

    data_shape_simulation = {}
    data_shape_means = {}
    for field in sorted(data_shape.model_fields):
        d = getattr(data_shape, field)
        if isinstance(d, Interval):
            data_shape_simulation[field] = interval_percentile(d, percentiles)
            if d.can_simulate:
                data_shape_means[field] = certain_float(d.mid)
            else:
                data_shape_means[field] = d
        else:
            data_shape_simulation[field] = [d] * len(percentiles)
            data_shape_means[field] = d

    results = []
    for i in range(len(percentiles)):
        try:
            query_pattern = QueryPattern(
                **{
                    f: query_pattern_simulation[f][i]
                    for f in sorted(query_pattern.model_fields)
                }
            )
        except Exception as exp:
            raise exp
        data_shape = DataShape(
            **{f: data_shape_simulation[f][i] for f in sorted(data_shape.model_fields)}
        )
        d = desires.model_copy(deep=True)
        d.query_pattern = query_pattern
        d.data_shape = data_shape
        results.append(d)

    mean_qp = QueryPattern(
        **{f: query_pattern_means[f] for f in sorted(query_pattern.model_fields)}
    )
    mean_ds = DataShape(
        **{f: data_shape_means[f] for f in sorted(data_shape.model_fields)}
    )
    d = desires.model_copy(deep=True)
    d.query_pattern = mean_qp
    d.data_shape = mean_ds

    return results, d


def _set_instance_objects(
    desires: CapacityDesires,
    hardware: Hardware,
):
    if desires.current_clusters:
        for zonal_cluster_capacity in desires.current_clusters.zonal:
            if zonal_cluster_capacity.cluster_instance_name in hardware.instances:
                zonal_cluster_capacity.cluster_instance = hardware.instances[
                    zonal_cluster_capacity.cluster_instance_name
                ]
            else:
                raise ValueError(
                    f"Model not trained to right size clusters that are of instance"
                    f" types {zonal_cluster_capacity.cluster_instance_name}"
                )
        for regional_cluster_capacity in desires.current_clusters.regional:
            if regional_cluster_capacity.cluster_instance_name in hardware.instances:
                regional_cluster_capacity.cluster_instance = hardware.instances[
                    regional_cluster_capacity.cluster_instance_name
                ]
            else:
                raise ValueError(
                    f"Model not trained to right size clusters that are of instance"
                    f" types {regional_cluster_capacity.cluster_instance_name}"
                )


def _allow_instance(
    instance: Instance,
    allowed_names: Sequence[str],
    allowed_lifecycles: Sequence[Lifecycle],
    allowed_platforms: Set[Platform],
) -> bool:
    # If the user has explicitly asked for particular families instead
    # of all lifecycles filter based on that
    if allowed_names:
        if instance.name not in allowed_names:
            if instance.family not in allowed_names:
                return False
    # Otherwise consider lifecycle (default) and platform
    else:
        if instance.lifecycle not in allowed_lifecycles:
            return False
        if allowed_platforms.isdisjoint(instance.platforms):
            return False

    return True


def _allow_drive(
    drive: Drive,
    allowed_names: Sequence[str],
    allowed_lifecycles: Sequence[Lifecycle],
    allowed_drives: Set[str],
) -> bool:
    # If the user has explicitly asked for particular families instead
    # of all lifecycles filter based on that
    if allowed_names:
        if drive.name not in allowed_names:
            return False
    # Otherwise consider lifecycle (default)
    else:
        if drive.lifecycle not in allowed_lifecycles:
            return False

    if drive.name not in allowed_drives:
        return False

    return True


def _regret(
    capacity_plans: Sequence[Tuple[CapacityDesires, CapacityPlan]],
    regret_params: CapacityRegretParameters,
    model: CapacityModel,
) -> Sequence[Tuple[CapacityPlan, CapacityDesires, float]]:
    plans_by_regret = []

    # Unfortunately has to be O(N^2) since regret isn't symmetric.
    # We could create the entire NxN regret matrix and use
    # einsum('ij->i') to quickly do a row wise sum, but that would
    # require a _lot_ more memory than this ...
    regret = np.zeros(len(capacity_plans), dtype=np.float64)
    for i, proposed_plan in enumerate(capacity_plans):
        for j, optimal_plan in enumerate(capacity_plans):
            if j == i:
                regret[j] = 0

            regret[j] = sum(
                model.regret(
                    regret_params=regret_params,
                    optimal_plan=optimal_plan[1],
                    proposed_plan=proposed_plan[1],
                ).values()
            )
        plans_by_regret.append(
            (proposed_plan[1], proposed_plan[0], np.einsum("i->", regret))
        )

    plans_by_regret.sort(key=lambda p: p[2])
    return plans_by_regret


def _add_requirement(requirement, accum):
    if requirement.requirement_type not in accum:
        accum[requirement.requirement_type] = {}

    requirements = accum[requirement.requirement_type]

    for field in sorted(requirement.model_fields):
        d = getattr(requirement, field)
        if isinstance(d, Interval):
            if field not in requirements:
                requirements[field] = [d]
            else:
                requirements[field].append(d)


def _merge_models(plans_by_model, zonal_requirements, regional_requirements):
    capacity_plans = []
    for composed in zip(*filter(lambda x: x, plans_by_model)):
        merged_plans = [functools.reduce(merge_plan, composed)]
        if len(merged_plans) == 0:
            continue

        capacity_plans.append(merged_plans[0])
        plan_requirements = merged_plans[0].requirements
        for req in plan_requirements.zonal:
            _add_requirement(req, zonal_requirements)
        for req in plan_requirements.regional:
            _add_requirement(req, regional_requirements)
    return capacity_plans


def _in_allowed(inp: str, allowed: Sequence[str]) -> bool:
    if not allowed:
        return True
    else:
        return inp in allowed


class CapacityPlanner:
    def __init__(
        self,
        default_num_simulations=128,
        default_num_results=2,
        default_lifecycles=(Lifecycle.stable, Lifecycle.beta),
    ):
        self._shapes: HardwareShapes = shapes
        self._models: Dict[str, CapacityModel] = {}

        self._default_num_simulations = default_num_simulations
        self._default_num_results = default_num_results
        self._default_regret_params = CapacityRegretParameters()
        self._default_lifecycles = default_lifecycles

    def register_group(self, group: Callable[[], Dict[str, CapacityModel]]):
        for name, model in group().items():
            self.register_model(name, model)

    def register_model(self, name: str, capacity_model: CapacityModel):
        self._models[name] = capacity_model

    @property
    def models(self) -> Dict[str, CapacityModel]:
        return self._models

    @property
    def hardware_shapes(self) -> HardwareShapes:
        return self._shapes

    def instance(self, name: str, region: Optional[str] = None) -> Instance:
        return self.hardware_shapes.instance(name, region=region)

    def _plan_percentiles(  # pylint: disable=too-many-positional-arguments
        self,
        model_name: str,
        percentiles: Tuple[int, ...],
        region: str,
        desires: CapacityDesires,
        lifecycles: Optional[Sequence[Lifecycle]] = None,
        instance_families: Optional[Sequence[str]] = None,
        drives: Optional[Sequence[str]] = None,
        num_results: Optional[int] = None,
        num_regions: int = 3,
        extra_model_arguments: Optional[Dict[str, Any]] = None,
    ) -> Tuple[Sequence[CapacityPlan], Dict[int, Sequence[CapacityPlan]]]:
        if model_name not in self._models:
            raise ValueError(
                f"model_name={model_name} does not exist. "
                f"Try {sorted(list(self._models.keys()))}"
            )

        extra_model_arguments = extra_model_arguments or {}
        lifecycles = lifecycles or self._default_lifecycles

        model_mean_desires: Dict[str, CapacityDesires] = {}
        sorted_percentiles = sorted(percentiles)
        model_percentile_desires: List[Dict[str, CapacityDesires]] = []
        for _ in sorted_percentiles:
            model_percentile_desires.append({})

        for sub_model, sub_desires in self._sub_models(
            model_name=model_name,
            desires=desires,
            extra_model_arguments=extra_model_arguments,
        ):
            percentile_inputs, mean_desires = model_desires_percentiles(
                desires=sub_desires, percentiles=sorted_percentiles
            )
            model_mean_desires[sub_model] = mean_desires
            index = 0
            for percentile_input in percentile_inputs:
                model_percentile_desires[index][sub_model] = percentile_input
                index = index + 1

        mean_plan = self._mean_plan(
            drives,
            extra_model_arguments,
            instance_families,
            lifecycles,
            num_regions,
            num_results,
            region,
            model_mean_desires,
        )
        percentile_plans = self._group_plans_by_percentile(
            drives,
            extra_model_arguments,
            instance_families,
            lifecycles,
            num_regions,
            num_results,
            region,
            model_percentile_desires,
            sorted_percentiles,
        )

        return mean_plan, percentile_plans

    def _group_plans_by_percentile(  # pylint: disable=too-many-positional-arguments
        self,
        drives,
        extra_model_arguments,
        instance_families,
        lifecycles,
        num_regions,
        num_results,
        region,
        model_percentile_desires,
        sorted_percentiles,
    ):
        percentile_plans = {}
        for index, percentile in enumerate(sorted_percentiles):
            percentile_plan = []
            for percentile_sub_model, percentile_sub_desire in model_percentile_desires[
                index
            ].items():
                percentile_sub_plan = self._plan_certain(
                    model_name=percentile_sub_model,
                    region=region,
                    desires=percentile_sub_desire,
                    num_results=num_results,
                    num_regions=num_regions,
                    extra_model_arguments=extra_model_arguments,
                    lifecycles=lifecycles,
                    instance_families=instance_families,
                    drives=drives,
                )
                if percentile_sub_plan:
                    percentile_plan.append(percentile_sub_plan)

            percentile_plans[percentile] = cast(
                Sequence[CapacityPlan],
                [
                    functools.reduce(merge_plan, composed)
                    for composed in zip(*percentile_plan)
                ],
            )
        return percentile_plans

    def _mean_plan(  # pylint: disable=too-many-positional-arguments
        self,
        drives,
        extra_model_arguments,
        instance_families,
        lifecycles,
        num_regions,
        num_results,
        region,
        model_mean_desires,
    ):
        mean_plans = []
        for mean_sub_model, mean_sub_desire in model_mean_desires.items():
            mean_sub_plan = self._plan_certain(
                model_name=mean_sub_model,
                region=region,
                desires=mean_sub_desire,
                num_results=num_results,
                num_regions=num_regions,
                extra_model_arguments=extra_model_arguments,
                lifecycles=lifecycles,
                instance_families=instance_families,
                drives=drives,
            )
            if mean_sub_plan:
                mean_plans.append(mean_sub_plan)
        mean_plan = cast(
            Sequence[CapacityPlan],
            [functools.reduce(merge_plan, composed) for composed in zip(*mean_plans)],
        )
        return mean_plan

    def plan_certain(  # pylint: disable=too-many-positional-arguments
        self,
        model_name: str,
        region: str,
        desires: CapacityDesires,
        lifecycles: Optional[Sequence[Lifecycle]] = None,
        instance_families: Optional[Sequence[str]] = None,
        drives: Optional[Sequence[str]] = None,
        num_results: Optional[int] = None,
        num_regions: int = 3,
        extra_model_arguments: Optional[Dict[str, Any]] = None,
    ) -> Sequence[CapacityPlan]:
        if model_name not in self._models:
            raise ValueError(
                f"model_name={model_name} does not exist. "
                f"Try {sorted(list(self._models.keys()))}"
            )

        extra_model_arguments = extra_model_arguments or {}
        lifecycles = lifecycles or self._default_lifecycles

        results = []

        for sub_model, sub_desires in self._sub_models(
            model_name=model_name,
            desires=desires,
            extra_model_arguments=extra_model_arguments,
        ):
            sub_plan = self._plan_certain(
                model_name=sub_model,
                region=region,
                desires=sub_desires,
                num_results=num_results,
                num_regions=num_regions,
                extra_model_arguments=extra_model_arguments,
                lifecycles=lifecycles,
                instance_families=instance_families,
                drives=drives,
            )
            if sub_plan:
                results.append(sub_plan)

        return [functools.reduce(merge_plan, composed) for composed in zip(*results)]

    def _plan_certain(  # pylint: disable=too-many-positional-arguments
        self,
        model_name: str,
        region: str,
        desires: CapacityDesires,
        num_results: Optional[int] = None,
        num_regions: int = 3,
        lifecycles: Optional[Sequence[Lifecycle]] = None,
        instance_families: Optional[Sequence[str]] = None,
        drives: Optional[Sequence[str]] = None,
        extra_model_arguments: Optional[Dict[str, Any]] = None,
    ) -> Sequence[CapacityPlan]:
        extra_model_arguments = extra_model_arguments or {}
        model = self._models[model_name]

        plans = []
        for instance, drive, context in self.generate_scenarios(
            model, region, desires, num_regions, lifecycles, instance_families, drives
        ):
            plan = model.capacity_plan(
                instance=instance,
                drive=drive,
                context=context,
                desires=desires,
                extra_model_arguments=extra_model_arguments,
            )
            if plan is not None:
                plans.append(plan)

        # lowest cost first
        plans.sort(key=lambda p: (p.rank, p.candidate_clusters.total_annual_cost))

        num_results = num_results or self._default_num_results
        return reduce_by_family(plans)[:num_results]

    # Calculates the minimum cpu, memory, and network requirements based on desires.
    def _per_instance_requirements(self, desires) -> Tuple[int, float]:

        # Applications often set fixed reservations of heap or OS memory
        per_instance_mem = (
            desires.data_shape.reserved_instance_app_mem_gib
            + desires.data_shape.reserved_instance_system_mem_gib
        )

        # Applications often require a minimum amount of true parallelism
        per_instance_cores = int(
            math.ceil(
                desires.query_pattern.estimated_read_parallelism.mid
                + desires.query_pattern.estimated_write_parallelism.mid
            )
        )

        current_capacity = (
            None
            if desires.current_clusters is None
            else (
                desires.current_clusters.zonal[0]
                if len(desires.current_clusters.zonal)
                else desires.current_clusters.regional[0]
            )
        )
        # Return early if we dont have current_capacity set.
        if current_capacity is None or current_capacity.cluster_instance is None:
            return (per_instance_cores, per_instance_mem)

        # Calculate memory requirements based on current capacity
        current_memory_utilization_gib = current_capacity.memory_utilization_gib.high
        per_instance_mem = max(per_instance_mem, current_memory_utilization_gib)

        return (per_instance_cores, per_instance_mem)

    def generate_scenarios(  # pylint: disable=too-many-positional-arguments
        self,
        model,
        region,
        desires,
        num_regions,
        lifecycles,
        instance_families,
        drives,
    ):
        lifecycles = lifecycles or self._default_lifecycles
        instance_families = instance_families or []
        drives = drives or []

        hardware = self._shapes.region(region)

        context = RegionContext(
            zones_in_region=hardware.zones_in_region,
            services={n: s.model_copy(deep=True) for n, s in hardware.services.items()},
            num_regions=num_regions,
        )

        allowed_platforms: Set[Platform] = set(model.allowed_platforms())
        allowed_drives: Set[str] = set(drives or [])
        for drive_name in model.allowed_cloud_drives():
            if drive_name is None:
                allowed_drives = set()
                break
            allowed_drives.add(drive_name)
        if len(allowed_drives) == 0:
            allowed_drives.update(hardware.drives.keys())

        # Set current instance object if exists
        _set_instance_objects(desires, hardware)

        # We should not even bother with shapes that don't meet the minimums
        (
            per_instance_cores,
            per_instance_mem,
        ) = self._per_instance_requirements(desires)

        if model.run_hardware_simulation():
            for instance in hardware.instances.values():
                if not _allow_instance(
                    instance, instance_families, lifecycles, allowed_platforms
                ):
                    continue

                # If the instance doesn't have enough vertical resources, pass on it
                if (
                    per_instance_mem > instance.ram_gib
                    or per_instance_cores > instance.cpu
                ):
                    continue

                for drive in hardware.drives.values():
                    if not _allow_drive(drive, drives, lifecycles, allowed_drives):
                        continue

                    yield instance, drive, context
        else:
            instance = Instance.get_managed_instance()
            drive = Drive.get_managed_drive()
            yield instance, drive, context

    # pylint: disable=too-many-locals
    def plan(  # pylint: disable=too-many-positional-arguments
        self,
        model_name: str,
        region: str,
        desires: CapacityDesires,
        percentiles: Tuple[int, ...] = (5, 50, 95),
        simulations: Optional[int] = None,
        num_results: Optional[int] = None,
        num_regions: int = 3,
        lifecycles: Optional[Sequence[Lifecycle]] = None,
        instance_families: Optional[Sequence[str]] = None,
        drives: Optional[Sequence[str]] = None,
        regret_params: Optional[CapacityRegretParameters] = None,
        extra_model_arguments: Optional[Dict[str, Any]] = None,
        explain: bool = False,
    ) -> UncertainCapacityPlan:
        extra_model_arguments = extra_model_arguments or {}

        if not all(0 <= p <= 100 for p in percentiles):
            raise ValueError("percentiles must be an integer in the range [0, 100]")
        if model_name not in self._models:
            raise ValueError(
                f"model_name={model_name} does not exist. "
                f"Try {sorted(list(self._models.keys()))}"
            )

        regret_params = regret_params or self._default_regret_params
        simulations = simulations or self._default_num_simulations
        num_results = num_results or self._default_num_results
        lifecycles = lifecycles or self._default_lifecycles

        # requirement types -> values
        zonal_requirements: Dict[str, Dict] = {}
        regional_requirements: Dict[str, Dict] = {}

        regret_clusters_by_model: Dict[
            str, Sequence[Tuple[CapacityPlan, CapacityDesires, float]]
        ] = {}
        desires_by_model: Dict[str, CapacityDesires] = {}
        for sub_model, sub_desires in self._sub_models(
            model_name=model_name,
            desires=desires,
            extra_model_arguments=extra_model_arguments,
        ):
            desires_by_model[sub_model] = sub_desires
            model_plans: List[Tuple[CapacityDesires, Sequence[CapacityPlan]]] = []
            for sim_desires in model_desires(sub_desires, simulations):
                model_plans.append(
                    (
                        sim_desires,
                        self._plan_certain(
                            model_name=sub_model,
                            region=region,
                            desires=sim_desires,
                            num_results=1,
                            num_regions=num_regions,
                            extra_model_arguments=extra_model_arguments,
                            lifecycles=lifecycles,
                            instance_families=instance_families,
                            drives=drives,
                        ),
                    )
                )
            regret_clusters_by_model[sub_model] = _regret(
                capacity_plans=[
                    (sim_desires, plan[0]) for sim_desires, plan in model_plans if plan
                ],
                regret_params=regret_params,
                model=self._models[sub_model],
            )

        # Now accumulate across the composed models and return the top N
        # by distinct hardware type
        least_regret = reduce_by_family(
            _merge_models(
                # First param is the actual plan which we care about
                [
                    [plan[0] for plan in component]
                    for component in regret_clusters_by_model.values()
                ],
                zonal_requirements,
                regional_requirements,
            )
        )[:num_results]

        low_p, high_p = sorted(percentiles)[0], sorted(percentiles)[-1]

        final_zonal = []
        final_regional = []
        for req_type, samples in zonal_requirements.items():
            req = CapacityRequirement(
                requirement_type=req_type,
                **{
                    k: interval(samples=[i.mid for i in v], low_p=low_p, high_p=high_p)
                    for k, v in samples.items()
                },
            )
            final_zonal.append(req)
        for req_type, samples in regional_requirements.items():
            req = CapacityRequirement(
                requirement_type=req_type,
                **{
                    k: interval(samples=[i.mid for i in v], low_p=low_p, high_p=high_p)
                    for k, v in samples.items()
                },
            )
            final_regional.append(req)

        final_requirement = Requirements(zonal=final_zonal, regional=final_regional)

        mean_plan, percentile_plans = self._plan_percentiles(
            model_name=model_name,
            percentiles=percentiles,
            region=region,
            desires=desires,
            extra_model_arguments=extra_model_arguments,
            num_regions=num_regions,
            instance_families=instance_families,
        )

        result = UncertainCapacityPlan(
            requirements=final_requirement,
            least_regret=least_regret,
            mean=mean_plan,
            percentiles=percentile_plans,
            explanation=PlanExplanation(
                regret_params=regret_params,
                desires_by_model={
                    model: desires.merge_with(
                        self._models[model].default_desires(
                            desires_by_model[model], extra_model_arguments
                        )
                    )
                    for model in regret_clusters_by_model
                },
            ),
        )
        if explain:
            result.explanation.regret_clusters_by_model = regret_clusters_by_model
            result.explanation.context["regret"] = least_regret

        return result

    def _sub_models(
        self,
        model_name: str,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ):
        queue: List[Tuple[CapacityDesires, str]] = [(desires, model_name)]
        models_used = []

        while queue:
            parent_desires, sub_model = queue.pop()
            # prevent infinite loop of models for now
            if sub_model in models_used:
                continue
            models_used.append(sub_model)

            sub_desires = parent_desires.merge_with(
                self._models[sub_model].default_desires(
                    parent_desires, extra_model_arguments
                )
            )

            # We might have to compose this model with others depending on
            # the user requirement
            queue.extend(
                [
                    (modify_child_desires(desires), child_model)
                    for child_model, modify_child_desires in self._models[
                        sub_model
                    ].compose_with(desires, extra_model_arguments)
                ]
            )

            yield sub_model, sub_desires


planner = CapacityPlanner()
planner.register_group(netflix.models)



from service_capacity_modeling.capacity_planner import planner
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import QueryPattern


uncertain_mid = CapacityDesires(
    service_tier=1,
    query_pattern=QueryPattern(
        estimated_read_per_second=Interval(
            low=1000, mid=10000, high=100000, confidence=0.98
        ),
        estimated_write_per_second=Interval(
            low=1000, mid=10000, high=100000, confidence=0.98
        ),
    ),
    data_shape=DataShape(
        estimated_state_size_gib=Interval(low=100, mid=500, high=1000, confidence=0.98),
    ),
)


def test_repeated_plans():
    results = []
    for _ in range(5):
        results.append(
            planner.plan(
                model_name="org.netflix.cassandra",
                region="us-east-1",
                desires=uncertain_mid,
            ).model_dump_json()
        )

    a = [hash(x) for x in results]
    # We should end up with consistent results
    assert all(i == a[0] for i in a)


def test_compositional():
    direct_result = planner.plan(
        model_name="org.netflix.cassandra",
        region="us-east-1",
        desires=uncertain_mid,
        num_results=4,
        explain=True,
    )
    composed_result = planner.plan(
        model_name="org.netflix.key-value",
        region="us-east-1",
        desires=uncertain_mid,
        num_results=4,
        explain=True,
    )

    count = len(direct_result.least_regret)
    direct_clusters = []
    for i in range(count):
        direct_cluster = direct_result.least_regret[i].candidate_clusters.zonal[0]
        direct_clusters.append(direct_cluster)
        # FIXME(josephl): It appears that since we are now zipping the
        # regional and zonal clusters we can get repeats in the zonal.
        # This is odd to me but not related to the 6th gen instances
        composed_cluster = composed_result.least_regret[i].candidate_clusters.zonal[0]
        assert composed_cluster in direct_clusters

        java = composed_result.least_regret[i].candidate_clusters.regional[0]
        assert java.cluster_type == "dgwkv"
        # usually like 15 * 4 = ~50
        assert 100 > java.count * java.instance.cpu > 20


def test_multiple_options():
    result = planner.plan(
        model_name="org.netflix.cassandra",
        region="us-east-1",
        desires=uncertain_mid,
        num_results=4,
        simulations=6,
    )
    least_regret = result.least_regret
    # With only 128 simulations we have < 4 instance families
    assert len(least_regret) < 4
    families = [lr.candidate_clusters.zonal[0].instance.family for lr in least_regret]
    for f in families:
        assert f in {"i3en", "m6id", "m5d", "i4i", "i3"}

    # With 1024 simulations we get a 4th instance family (r5)
    result = planner.plan(
        model_name="org.netflix.cassandra",
        region="us-east-1",
        desires=uncertain_mid,
        num_results=4,
        simulations=1024,
    )
    least_regret = result.least_regret
    assert len(least_regret) == 4

    families = [lr.candidate_clusters.zonal[0].instance.family for lr in least_regret]
    for f in families:
        assert f.startswith("i") or f.startswith("r") or f.startswith("m")



# pylint: disable=too-many-lines
from __future__ import annotations

import re
import sys
from decimal import Decimal
from enum import Enum
from fractions import Fraction
from functools import lru_cache
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np
from pydantic import BaseModel
from pydantic import computed_field
from pydantic import ConfigDict
from pydantic import Field

GIB_IN_BYTES = 1024 * 1024 * 1024
MIB_IN_BYTES = 1024 * 1024
MEGABIT_IN_BYTES = (1000 * 1000) / 8


class ExcludeUnsetModel(BaseModel):
    def model_dump(self, *args, **kwargs):
        if "exclude_unset" not in kwargs:
            kwargs["exclude_unset"] = True
        return super().model_dump(*args, **kwargs)

    def model_dump_json(self, *args, **kwargs):
        if "exclude_unset" not in kwargs:
            kwargs["exclude_unset"] = True
        return super().model_dump_json(*args, **kwargs)


###############################################################################
#              Models (structs) for how we describe intervals                 #
###############################################################################


class IntervalModel(str, Enum):
    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f"D({self.value})"

    gamma = "gamma"
    beta = "beta"


class Interval(ExcludeUnsetModel):
    low: float
    mid: float
    high: float
    # How confident are we of this interval
    confidence: float = 1.0
    # How to approximate this interval (e.g. with a beta distribution)
    model_with: IntervalModel = IntervalModel.beta
    # If we should allow simulation of this interval, some models might not
    # be able to simulate or some properties might not want to
    allow_simulate: bool = True

    minimum_value: Optional[float] = None
    maximum_value: Optional[float] = None
    model_config = ConfigDict(frozen=True, protected_namespaces=())

    @property
    def can_simulate(self):
        return self.confidence <= 0.99 and self.allow_simulate

    @property
    def minimum(self):
        if self.minimum_value is None:
            if self.confidence == 1.0:
                return self.low * 0.999
            return self.low / 2

        return self.minimum_value

    @property
    def maximum(self):
        if self.maximum_value is None:
            if self.confidence == 1.0:
                return self.high * 1.001
            return self.high * 2
        return self.maximum_value

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def scale(self, factor: float) -> Interval:
        minimum_value = (
            self.minimum * factor if self.minimum_value is not None else None
        )
        maximum_value = (
            self.maximum * factor if self.maximum_value is not None else None
        )
        return Interval(
            low=self.low * factor,
            mid=self.mid * factor,
            high=self.high * factor,
            confidence=self.confidence,
            model_with=self.model_with,
            allow_simulate=self.allow_simulate,
            minimum_value=minimum_value,
            maximum_value=maximum_value,
        )

    def offset(self, delta: float) -> Interval:
        minimum_value = self.minimum + delta if self.minimum_value is not None else None
        maximum_value = self.maximum + delta if self.maximum_value is not None else None
        return Interval(
            low=self.low + delta,
            mid=self.mid + delta,
            high=self.high + delta,
            confidence=self.confidence,
            model_with=self.model_with,
            allow_simulate=self.allow_simulate,
            minimum_value=minimum_value,
            maximum_value=maximum_value,
        )


class FixedInterval(Interval):
    allow_simulate: bool = False


@lru_cache(2048)
def certain_int(x: int) -> Interval:
    return Interval(low=x, mid=x, high=x, confidence=1.0)


@lru_cache(2048)
def certain_float(x: float) -> Interval:
    return Interval(low=x, mid=x, high=x, confidence=1.0)


@lru_cache(2048)
def fixed_float(x: float) -> FixedInterval:
    return FixedInterval(low=x, mid=x, high=x, confidence=1.0)


def interval(samples: Sequence[float], low_p: int = 5, high_p: int = 95) -> Interval:
    p = np.percentile(a=samples, q=[0, low_p, 50, high_p, 100])
    conf = (high_p - low_p) / 100
    return Interval(
        low=p[1],
        mid=p[2],
        high=p[3],
        minimum_value=p[0],
        maximum_value=p[4],
        confidence=conf,
    )


def normalized_aws_size(name: str) -> Fraction:
    """Normalizes an AWS shape to a fractional xlarge unit"""
    _, size = name.split(".")
    numeric = re.findall(r"\d+", size)
    if numeric:
        assert len(numeric) == 1
        return Fraction(float(numeric[0]))
    return {
        "small": Fraction(1, 8),
        "medium": Fraction(1, 4),
        "large": Fraction(1, 2),
        "xlarge": Fraction(1),
        # Is this always true?
        "metal": Fraction(48, 1),
    }[size]


###############################################################################
#              Models (structs) for how we describe hardware                  #
###############################################################################


class Lifecycle(str, Enum):
    """Represents the lifecycle of hardware from initial preview
    to end-of-life.

    For example a particular shape of hardware might be released under
    preview and adventurous models may wish to use those, and then
    more risk-averse workloads may wait for stability.

    By default models are shown beta and stable shapes to pick from
    """

    alpha = "alpha"
    beta = "beta"
    stable = "stable"
    deprecated = "deprecated"
    end_of_life = "end-of-life"


class DriveType(str, Enum):
    """Represents the type of drive"""

    local_ssd = "local-ssd"
    local_hdd = "local-hdd"
    attached_ssd = "attached-ssd"
    attached_hdd = "attached-hhd"


class Drive(ExcludeUnsetModel):
    """Represents a cloud drive e.g. EBS or ephemeral drives

    This model is generic to any cloud
    """

    name: str
    drive_type: DriveType = DriveType.local_ssd
    size_gib: int = 0
    read_io_per_s: Optional[int] = None
    write_io_per_s: Optional[int] = None
    throughput: Optional[int] = None
    # If this drive has single tenant IO capacity, for example a single
    # physical drive versus a virtualised drive
    single_tenant: bool = True
    # If this drive can scale, how large can it scale to
    max_scale_size_gib: int = 0
    # If this drive can scale IO, how large can it scale to
    max_scale_io_per_s: int = 0
    # How large is an "IO" against this device
    block_size_kib: int = 4
    # When sequential how much IO is grouped into a single "IO"
    # https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/ebs-io-characteristics.html
    # Some cloud drives can group sequential ops together and DBs take advantage
    group_size_kib: int = 4

    lifecycle: Lifecycle = Lifecycle.stable
    compatible_families: List[str] = []

    annual_cost_per_gib: float = 0
    # Tuples of [max_size, annual cost]
    # [32000, 0.78], ...
    annual_cost_per_read_io: List[Tuple[float, float]] = []
    annual_cost_per_write_io: List[Tuple[float, float]] = []

    # These defaults are assuming a cloud SSD like a gp2 volume
    # If you disagree please change them in your hardware description
    read_io_latency_ms: FixedInterval = FixedInterval(
        low=0.8, mid=1, high=2, confidence=0.9
    )
    write_io_latency_ms: FixedInterval = FixedInterval(
        low=0.6, mid=2, high=3, confidence=0.9
    )

    @property
    def rand_io_size_kib(self) -> int:
        return self.block_size_kib

    @property
    def seq_io_size_kib(self) -> int:
        return max(self.block_size_kib, self.group_size_kib)

    @property
    def max_size_gib(self):
        if self.max_scale_size_gib != 0:
            return self.max_scale_size_gib
        else:
            return self.size_gib

    @property
    def max_io_per_s(self):
        if self.max_scale_io_per_s != 0:
            return self.max_scale_io_per_s
        else:
            return sys.maxsize

    @computed_field(return_type=float)  # type: ignore
    @property
    def annual_cost(self):
        size = self.size_gib or 0
        r_ios = self.read_io_per_s or 0
        w_ios = self.write_io_per_s or 0

        # Time to do income taxes ...
        # Inputs are ranges of io limits and costs for ios in that range
        # [(32000.0, 0.78),
        #  (64000.0, 0.552),
        #  (160000.0, 0.384)]
        r_cost, w_cost, offset = 0.0, 0.0, 0.0
        if self.annual_cost_per_read_io:
            for end, cost in self.annual_cost_per_read_io:
                charge_ios = min(r_ios, end) - offset
                r_cost += charge_ios * cost
                offset += charge_ios
                if offset >= r_ios:
                    break

        offset = 0.0
        if self.annual_cost_per_write_io:
            for end, cost in self.annual_cost_per_write_io:
                charge_ios = min(w_ios, end) - offset
                w_cost += charge_ios * cost
                offset += charge_ios
                if offset >= w_ios:
                    break

        return size * self.annual_cost_per_gib + r_cost + w_cost

    @staticmethod
    def get_managed_drive() -> Drive:
        return Drive(name="managed")


class Platform(str, Enum):
    """Represents the platform of the hardware

    For example a particular hardware type might offer x86_64, arm, or be a managed
    instance type that only works with managed RDBMS like Aurora Postgres.
    """

    # Most Intel and AMD instance types
    amd64 = "amd64"
    # Graviton and other ARM based instance types
    arm64 = "arm64"
    # Special purpose aurora type
    aurora_mysql = "Aurora MySQL"
    # Special purpose aurora type
    aurora_postgres = "Aurora PostgreSQL"


class Instance(ExcludeUnsetModel):
    """Represents a cloud instance aka Hardware Shape

    This model is generic to any cloud.
    """

    name: str
    cpu: int = Field(
        title="CPU threads (vCPUs or effective cores)",
        description=(
            "The number of CPU threads this instance has, typically a SMT thread "
            "but if equal to Instance.cpu_cores, actual cores."
        ),
    )
    cpu_cores: Optional[int] = Field(
        default=None,
        title="If known the number of physical cores present on the machine",
        description=(
            "Most Instance.cpu are really threads, meaning they are not truly "
            "parallel. Some Instance.cpu are full fat cores, in which case this "
            "will be set and be equal to the cpu count. Use the cores getter in code"
        ),
    )
    cpu_ghz: float = Field(
        title="The clock frequency of these cores",
        description=(
            "How fast the base clock of the cores are. We assume only baseline"
            " all-core sustained turbo for the purpose of capacity planning."
        ),
    )
    cpu_ipc_scale: float = Field(
        default=1.0,
        title="Instruction per clock scale: core speed is multiplied by this",
        description=(
            "Not all cores or ghz are created equal, and if we want to represent that "
            "a CPU is more productive per unit frequency, this fractional scaling "
            "influences core normalization logic. Note that in the longer term we "
            "expect to replace this with a standard measure of average performance."
        ),
    )
    ram_gib: float
    net_mbps: float
    drive: Optional[Drive] = None
    annual_cost: float = 0
    lifecycle: Lifecycle = Lifecycle.stable
    # Typically hardware has a single platform, but sometimes they can act in multiple
    platforms: List[Platform] = [Platform.amd64]

    family_separator: str = "."

    @property
    def family(self):
        return self.name.rsplit(self.family_separator, 1)[0]

    @property
    def size(self):
        return self.name.rsplit(self.family_separator, 1)[1]

    @property
    def cores(self):
        if self.cpu_cores is not None:
            return self.cpu_cores
        return self.cpu // 2

    @staticmethod
    def get_managed_instance() -> Instance:
        return Instance(
            name="managed.0", cpu=0, cpu_ghz=0, ram_gib=0, net_mbps=0, drive=None
        )

    def merge_with(self, overrides: "Instance") -> "Instance":
        self_dict = self.model_dump()
        other_dict = overrides.model_dump(exclude_unset=True)

        for (k, v) in other_dict.items():
            # TODO we need a deep merge on drive (recursive merge)
            if k in ("platforms",):
                # Unique merge platforms
                merged_platforms = list(self_dict.get("platforms", []))
                merged_platforms = merged_platforms + [
                    i
                    for i in other_dict.get("platforms", [])
                    if i not in merged_platforms
                ]
                self_dict["platforms"] = merged_platforms
            else:
                self_dict[k] = v
        return Instance(**self_dict)


default_reference_shape = Instance(
    name="default_reference_shape",
    cpu=4,
    # Much of our benchmarking was carried out in the 5th gen i2 shape, can
    # adjust this up once we go through and fix all the latency distributions
    cpu_ghz=2.3,
    cpu_ipc_scale=1.0,
    ram_gib=8.0,
    net_mbps=1000,
)


class Service(ExcludeUnsetModel):
    """Represents a cloud service, such as a blob store (S3) or
    managed service such as DynamoDB or RDS.

    This model is generic to any cloud.
    """

    name: str
    size_gib: int = 0

    annual_cost_per_gib: Union[float, List[Tuple[float, float]]] = 0
    annual_cost_per_read_io: float = 0
    annual_cost_per_write_io: float = 0
    annual_cost_per_core: float = 0

    # These defaults assume a cloud blob storage like S3
    read_io_latency_ms: FixedInterval = FixedInterval(
        low=1, mid=5, high=50, confidence=0.9
    )
    write_io_latency_ms: FixedInterval = FixedInterval(
        low=1, mid=10, high=50, confidence=0.9
    )

    def annual_cost_gib(self, data_gib: float = 0):
        if isinstance(self.annual_cost_per_gib, float):
            return self.annual_cost_per_gib * data_gib
        else:
            _annual_data = data_gib
            transfer_costs = list(self.annual_cost_per_gib)
            annual_cost = 0.0
            for transfer_cost in transfer_costs:
                if _annual_data <= 0:
                    break
                if transfer_cost[0] > 0:
                    annual_cost += (
                        min(_annual_data, transfer_cost[0]) * transfer_cost[1]
                    )
                    _annual_data -= transfer_cost[0]
                else:
                    # final remaining data transfer cost
                    annual_cost += _annual_data * transfer_cost[1]
        return annual_cost


class RegionContext(ExcludeUnsetModel):
    services: Dict[str, Service] = {}
    zones_in_region: int = 3
    num_regions: int = 3


class Hardware(ExcludeUnsetModel):
    """Represents a hardware deployment

    In EC2 this maps to:
        instances: instance type -> Instance(cpu, mem, cost, etc...)
        drives: ebs type -> Drive(cost per _GiB year_, etc...)
        services: service type -> Service(name, params, cost, etc ...)
    """

    # How many zones of compute exist in this region of compute
    zones_in_region: int = 3
    # Per instance shape information e.g. cpu, ram, cpu etc ...
    instances: Dict[str, Instance] = {}
    # Per drive type information and cost
    drives: Dict[str, Drive] = {}
    # Per service information and cost
    services: Dict[str, Service] = {}


class GlobalHardware(ExcludeUnsetModel):
    """Represents all possible hardware shapes in all regions

    In EC2 this maps to:
        us-east-1 -> Hardware available in us-east-1
        us-east-2 -> Hardware available in us-east-2
        ...
    """

    # Per region hardware shapes
    regions: Dict[str, Hardware]


class InstancePricing(ExcludeUnsetModel):
    annual_cost: float = 0
    lifecycle: Optional[Lifecycle] = None


class DrivePricing(ExcludeUnsetModel):
    annual_cost_per_gib: float = 0
    annual_cost_per_read_io: List[Tuple[float, float]] = []
    annual_cost_per_write_io: List[Tuple[float, float]] = []


class ServicePricing(ExcludeUnsetModel):
    annual_cost_per_gib: Union[float, List[Tuple[float, float]]] = 0
    annual_cost_per_read_io: float = 0
    annual_cost_per_write_io: float = 0
    annual_cost_per_core: float = 0


class HardwarePricing(ExcludeUnsetModel):
    instances: Dict[str, InstancePricing]
    drives: Dict[str, DrivePricing]
    services: Dict[str, ServicePricing]
    zones_in_region: int = 3


class Pricing(ExcludeUnsetModel):
    regions: Dict[str, HardwarePricing]


###############################################################################
#               Models (structs) for how we plan capacity                     #
###############################################################################


class AccessPattern(str, Enum):
    latency = "latency"
    throughput = "throughput"


class AccessConsistency(str, Enum):
    """See https://jepsen.io/consistency

    Generally speaking consistency is expensive, so models need to know what
    kind of consistency will be required in order to estimate CPU usage
    within a factor of 4-5x correctly.
    """

    # You cannot read writes ever
    never = "never"

    #
    # Single item consistency (most services)
    #

    # Best Effort: we might lose writes or reads might be stale or missing.
    #              Most caches offer this level of consistency.
    # Eventual: We will eventually reflect the latest successful write but
    #           there is some (often large) time bound on that eventuality.
    # Read-Your-Writes: The first "consistent" offering.
    best_effort = "best-effort"
    eventual = "eventual"
    read_your_writes = "read-your-writes"
    # Fully lineralizable, writes and reads
    linearizable = "linearizable"
    # Writes are linerizable but stale reads are possible (e.g. ZK)
    linearizable_stale = "linearizable-stale"

    #
    # Multiple item consistency (often "transactional" or "acid" services)
    #

    # All operations are serializable.
    # (e.g. CRDB in default settings)
    serializable = "serializable"
    # Writes are serializable but stale reads are possible
    # (e.g. CRDB with stale reads enabled, MySQL with read replicas, etc ...)
    serializable_stale = "serializable-stale"


AVG_ITEM_SIZE_BYTES: int = 1024


class Consistency(ExcludeUnsetModel):
    target_consistency: Optional[AccessConsistency] = Field(
        None,
        title="Consistency requirement on access",
        description=(
            "Stronger consistency access is generally more expensive."
            " The words used here to describe consistency attempt to "
            " align with the Jepsen tree of multi/single object "
            " consistency models: https://jepsen.io/consistency"
        ),
    )
    staleness_slo_sec: FixedInterval = Field(
        FixedInterval(low=0, mid=10, high=60),
        title="When stale reads are permitted what is the staleness requirement",
        description=(
            "Eventual consistency (aka stale reads) is usually bounded by some"
            " amount of time. Applications can use this to try to enforce when "
            " a write is available for reads"
        ),
    )


class GlobalConsistency(ExcludeUnsetModel):
    same_region: Consistency = Consistency(
        target_consistency=None,
        staleness_slo_sec=FixedInterval(low=0, mid=0.1, high=1),
    )
    cross_region: Consistency = Consistency(
        target_consistency=None,
        staleness_slo_sec=FixedInterval(low=10, mid=60, high=600),
    )


class QueryPattern(ExcludeUnsetModel):
    # Will the service primarily be accessed in a latency sensitive mode
    # (aka we care about P99) or throughput (we care about averages)
    access_pattern: AccessPattern = AccessPattern.latency
    access_consistency: GlobalConsistency = GlobalConsistency()

    # A main input, how many requests per second will we handle
    # We assume this is the mean of a range of possible outcomes
    estimated_read_per_second: Interval = certain_int(0)
    estimated_write_per_second: Interval = certain_int(0)

    # A main input, how much _on cpu_ time per operation do you take.
    # This depends heavily on workload, but this is a generally ok default
    # For a Java app (C or C++ will generally be about 10x better,
    # python 2-4x slower, etc...)
    estimated_mean_read_latency_ms: Interval = certain_float(1)
    estimated_mean_write_latency_ms: Interval = certain_float(1)

    # For stateful services the amount of data accessed per
    # read and write impacts disk and network provisioniong
    # For stateless services it mostly just impacts memory and network
    estimated_mean_read_size_bytes: Interval = certain_int(AVG_ITEM_SIZE_BYTES)
    estimated_mean_write_size_bytes: Interval = certain_int(AVG_ITEM_SIZE_BYTES // 2)

    # For workloads which have bursts of async work, what is the
    # expected parallelism of those workloads. Note the summation of
    # read and write parallelism will lower bound the number of cores.
    estimated_read_parallelism: Interval = Field(
        certain_int(1),
        title="Estimated per instance parallelism on read operations",
        description=(
            "The estimated amount of parallel work streams on a single "
            "host. For example a read triggers async callbacks that need "
            "to be executed truly in parallel (not just concurrent)."
        ),
    )
    estimated_write_parallelism: Interval = Field(
        certain_int(1),
        title="Estimated per instance parallelism on write operations",
        description=(
            "The estimated amount of parallel work streams on a single "
            "host. For example a write triggers async fanouts that need "
            "to be executed truly in parallel (not just concurrent)."
        ),
    )

    # The latencies at which oncall engineers get involved. We want
    # to provision such that we don't involve oncall
    # Note that these summary statistics will be used to create reasonable
    # distribution approximations of these operations (yielding p25, p99, etc)
    read_latency_slo_ms: FixedInterval = FixedInterval(
        low=0.4, mid=4, high=10, confidence=0.98
    )
    write_latency_slo_ms: FixedInterval = FixedInterval(
        low=0.4, mid=4, high=10, confidence=0.98
    )


class DataShape(ExcludeUnsetModel):
    estimated_state_size_gib: Interval = Field(
        certain_int(0),
        title="Estimated amount of state in the system in GiB",
        description=(
            "The estimated amount of state that will be stored."
            " Note that this is an estimate and doesn't need to be exact"
        ),
    )
    estimated_state_item_count: Optional[Interval] = None

    estimated_working_set_percent: Optional[Interval] = Field(
        None,
        title="Estimated working set percentage",
        description=(
            "The estimated percentage of data that will be accessed frequently"
            " and therefore must be kept hot in memory (e.g. 0.10). Note that "
            " models will generally estimate this from the latency SLO and "
            "latency model of the drives being attached"
        ),
    )

    # How compressible is this dataset. Note that databases might offer
    # better or worse compression strategies that will impact this
    #   Note that the ratio here is the forward ratio, e.g.
    #   A ratio of 2 means 2:1 compression (0.5 on disk size)
    #   A ratio of 5 means 5:1 compression (0.2 on disk size)
    estimated_compression_ratio: Interval = certain_float(1)

    # How much fixed memory must be provisioned per instance for the
    # application (e.g. for process heap memory)
    reserved_instance_app_mem_gib: float = 2

    # How much fixed memory must be provisioned per instance for the
    # system (e.g. for kernel and other system processes)
    reserved_instance_system_mem_gib: float = 1

    # How durable does this dataset need to be. We want to provision
    # sufficient replication and backups of data to achieve the target
    # durability SLO so we don't lose our customer's data. Note that
    # This is measured in orders of magnitude. So
    #   1000   = 1 - (1/1000) = 0.999
    #   10000  = 1 - (1/10000) = 0.9999
    durability_slo_order: FixedInterval = FixedInterval(
        low=1000, mid=10000, high=100000, confidence=0.98
    )


class CurrentClusterCapacity(ExcludeUnsetModel):
    cluster_instance_name: str
    cluster_instance: Optional[Instance] = None
    cluster_instance_count: Interval
    # The distribution cpu utilization in the cluster.
    cpu_utilization: Interval = certain_float(0.0)
    # The per node distribution of memory used in gib.
    memory_utilization_gib: Interval = certain_float(0.0)
    # The per node distribution of network used in mbps.
    network_utilization_mbps: Interval = certain_float(0.0)
    # The per node distribution of disk used in gib.
    disk_utilization_gib: Interval = certain_float(0.0)


# For services that are provisioned by zone (e.g. Cassandra, EVCache)
class CurrentZoneClusterCapacity(CurrentClusterCapacity):
    pass


# For services that are provisioned regionally (e.g. Java services, RDS, etc ..)
class CurrentRegionClusterCapacity(CurrentClusterCapacity):
    pass


class CurrentClusters(ExcludeUnsetModel):
    zonal: Sequence[CurrentZoneClusterCapacity] = []
    regional: Sequence[CurrentRegionClusterCapacity] = []
    services: Sequence[ServiceCapacity] = []


class BufferComponent(str, Enum):
    """Represents well known buffer components such as compute and storage

    Note that while these are common and defined here for models to share,
    models can have w.e. buffers they want, and this type should not enter
    the Buffers interface itself (should be str).
    """

    # [Query Pattern] a.k.a. "Traffic" related buffers, e.g. CPU and Network
    compute = "compute"
    # [Data Shape]    a.k.a. "Dataset" related buffers, e.g. Disk and Memory
    storage = "storage"

    # Resource specific component
    cpu = "cpu"
    network = "network"
    disk = "disk"
    memory = "memory"


class BufferIntent(str, Enum):
    # Most buffers show "desired" buffer, this is the default
    desired = "desired"
    # ratio on top of existing buffers to ensure exists. Generally combined
    # with a different desired buffer to ensure we don't just scale needlessly
    scale = "scale"
    # Ignore model preferences, just preserve existing buffers
    preserve = "preserve"


class Buffer(ExcludeUnsetModel):
    # The value of the buffer expressed as a ratio over "normal" load e.g. 1.5x
    ratio: float = 1.0
    # What is the intent of this buffer directive, almost always is desired
    intent: BufferIntent = BufferIntent.desired
    # The components of buffer this influences, almost always is "compute" (IPC success)
    components: List[str] = [BufferComponent.compute]
    # If this buffer was made up of other buffers, what contributed
    sources: Dict[str, Buffer] = {}


class Buffers(ExcludeUnsetModel):
    """Typical buffers (headroom) over the requirements to build into the system

    Note that typically callers make buffer choices based on business context for
    example a tier 1 service may request:

    Buffers(
        desired={
            "compute": Buffer(ratio: 1.5),
        }
    )

    And then models layer in their buffers, for example if a workload
    requires 10 CPU cores, but the operator of that workload  likes to build in
    2x buffer for background work (20 cores provisioned), they would express that
    as a model desire default of:

    Buffers(
        desired={
            "background": Buffer(ratio: 2.0, components=[BufferComponent.cpu]),
        }
    )

    In this case when we query the buffer_for_components(components=["cpu"])
    we get 2.0 * 1.5 = 3x because the model has reserved 2x for background work
    and the caller has asked for 1.5x load buffer - we also see both component
    buffers in the Buffer.source field.
    """

    # The default buffer if a specific buffer isn't known
    # Models should prefer to document their precise buffers in desired
    default: Buffer = Buffer(ratio=1.5)
    # Desired compute, storage, cpu, memory, etc... buffers
    desired: Dict[str, Buffer] = {}
    # Derive these buffers from current clusters or model context
    # Buffer.intent MUST be set on these Buffers to something other than "desired":
    #   scale    = ratio on top of existing buffers to ensure. Let the "derived"
    #              buffer multiplied by this ratio "needed". If the "needed" buffer
    #              is greater than desired, the needed buffer is created. If the
    #              "needed" buffer is less than desired, the needed buffer is created.
    #   preserve = ignore desired buffer entirely, just maintain existing buffers
    derived: Dict[str, Buffer] = {}


class CapacityDesires(ExcludeUnsetModel):
    # How critical is this cluster, impacts how much "extra" we provision
    # 0 = Critical to the product            (Product does not function)
    # 1 = Important to product with fallback (User experience degraded)
    # 2 = Care about it but don't wake up    (Internal apps)
    # 3 = Do not care                        (Testing)
    service_tier: int = 1

    # How will the service be queried
    query_pattern: QueryPattern = QueryPattern()

    # What will the state look like that is being queries
    data_shape: DataShape = DataShape()

    # What is the current deployment and utilization of the system
    current_clusters: Optional[CurrentClusters] = None

    # What are the desired buffers (headroom) state, mostly injected by models
    # Note if you pass current_clusters you can express "I need the status quo"
    # by setting the buffers you want to preserve in buffers.derived
    buffers: Buffers = Buffers()

    @property
    def reference_shape(self) -> Instance:
        # TODO: this should use the shape from current clusters if it is there
        return default_reference_shape

    def merge_with(self, defaults: "CapacityDesires") -> "CapacityDesires":
        # Now merge with the models default
        desires_dict = self.model_dump()
        default_dict = defaults.model_dump()

        default_dict.get("query_pattern", {}).update(
            desires_dict.pop("query_pattern", {})
        )
        default_dict.get("data_shape", {}).update(desires_dict.pop("data_shape", {}))

        # Buffers has deep structure we want to deep merge on
        if "buffers" not in default_dict:
            default_dict["buffers"] = {}
        default_buffers = default_dict["buffers"]
        desired_buffers = desires_dict.pop("buffers", {})
        if "default" in desired_buffers:
            default_buffers["default"] = desired_buffers["default"]
        for k, v in desired_buffers.get("desired", {}).items():
            default_buffers["desired"][k] = v

        default_buffers.setdefault("derived", {})
        for k, v in desired_buffers.get("derived", {}).items():
            default_buffers["derived"][k] = v

        default_dict.update(desires_dict)

        desires = CapacityDesires(**default_dict)

        # If user gave state item count but not size or size but not count
        # calculate the missing one from the other
        user_size = (
            self.model_dump()
            .get("data_shape", {})
            .get("estimated_state_size_gib", None)
        )
        user_count = self.data_shape.estimated_state_item_count
        item_size_bytes = desires.query_pattern.estimated_mean_write_size_bytes.mid
        if user_size is None and user_count is not None:
            desires.data_shape.estimated_state_size_gib = user_count.scale(
                factor=(item_size_bytes / GIB_IN_BYTES)
            )
        elif user_size is not None and user_count is None:
            user_size_gib = self.data_shape.estimated_state_size_gib
            desires.data_shape.estimated_state_item_count = user_size_gib.scale(
                factor=(GIB_IN_BYTES / item_size_bytes)
            )

        return desires


class CapacityRequirement(ExcludeUnsetModel):
    requirement_type: str

    # cpu_cores was calculated relative to this reference, mostly for
    # comparing latency to clock frequency
    reference_shape: Instance = default_reference_shape
    cpu_cores: Interval
    mem_gib: Interval = certain_int(0)
    network_mbps: Interval = certain_int(0)
    disk_gib: Interval = certain_int(0)

    context: Dict = {}


class ClusterCapacity(ExcludeUnsetModel):
    cluster_type: str

    count: int
    instance: Instance
    attached_drives: Sequence[Drive] = ()
    annual_cost: float
    # When provisioning services we might need to signal they
    # should have certain configuration, for example flags that
    # affect durability shut off
    cluster_params: Dict = {}


class ServiceCapacity(ExcludeUnsetModel):
    service_type: str
    annual_cost: float
    # If this cost should impact regret. Usually they do not since they
    # are proportional to the input desire not the output cluster
    regret_cost: bool = False
    # Often while provisioning cloud services we need to represent
    # parameters to the cloud APIs, use this to inject those from models
    service_params: Dict = {}


# For services that are provisioned by zone (e.g. Cassandra, EVCache)
class ZoneClusterCapacity(ClusterCapacity):
    pass


# For services that are provisioned regionally (e.g. Java services, RDS, etc ..)
class RegionClusterCapacity(ClusterCapacity):
    pass


class Requirements(ExcludeUnsetModel):
    zonal: Sequence[CapacityRequirement] = []
    regional: Sequence[CapacityRequirement] = []

    # Commonly a model regrets "spend", lack of "disk" space and sometimes
    # lack of "mem"ory. Default options are ["cost", "disk", "mem"]
    regrets: Sequence[str] = ("spend", "disk")

    # Used by models to have custom regret components
    # pylint: disable=unused-argument
    @staticmethod
    def regret(
        name: str, optimal_plan: "CapacityPlan", proposed_plan: "CapacityPlan"
    ) -> float:
        return 0.0


class Clusters(ExcludeUnsetModel):
    annual_costs: Dict[str, Decimal]
    zonal: Sequence[ZoneClusterCapacity] = []
    regional: Sequence[RegionClusterCapacity] = []
    services: Sequence[ServiceCapacity] = []

    # Backwards compatibility for total_annual_cost
    @computed_field(return_type=float)  # type: ignore
    @property
    def total_annual_cost(self) -> float:
        return round(float(sum(self.annual_costs.values())), 2)


class CapacityPlan(ExcludeUnsetModel):
    requirements: Requirements
    candidate_clusters: Clusters
    rank: int = 0


# Parameters to cost functions of the form
# let y = the optimal value
# let x = the proposed value
# let cost = (x - y) ^ exponent
class Regret(ExcludeUnsetModel):
    over_provision_cost: float = 0
    under_provision_cost: float = 0
    exponent: float = 1.0


class CapacityRegretParameters(ExcludeUnsetModel):
    # How much do we regret spending too much or too little money
    spend: Regret = Regret(
        over_provision_cost=1, under_provision_cost=1.25, exponent=1.2
    )

    # For every GiB we are underprovisioned by default cost $1 / year / GiB
    disk: Regret = Regret(
        over_provision_cost=0, under_provision_cost=1.1, exponent=1.05
    )

    # For every GiB we are underprovisioned on memory (for datastores
    # storing data in RAM), regret under_provisioning slightly more than disk
    mem: Regret = Regret(over_provision_cost=0, under_provision_cost=1.5, exponent=1.1)

    # Any additional metric we want to regret from the models
    # just have to be returned in the requirement
    extra: Dict[str, Regret] = {}


class PlanExplanation(ExcludeUnsetModel):
    regret_params: CapacityRegretParameters
    regret_clusters_by_model: Dict[
        str, Sequence[Tuple[CapacityPlan, CapacityDesires, float]]
    ] = {}
    desires_by_model: Dict[str, CapacityDesires] = {}
    context: Dict[str, Any] = {}


class UncertainCapacityPlan(ExcludeUnsetModel):
    requirements: Requirements
    least_regret: Sequence[CapacityPlan]
    mean: Sequence[CapacityPlan]
    percentiles: Dict[int, Sequence[CapacityPlan]]
    explanation: PlanExplanation



from service_capacity_modeling.hardware import shapes

io2 = shapes.hardware.regions["us-east-1"].drives["io2"]


def test_io2_pricing():
    d = io2.model_copy()
    d.size_gib = 100
    d.read_io_per_s = 33000
    assert d.annual_cost == 25662

    d.read_io_per_s = 31000
    d.write_io_per_s = 34000
    assert d.annual_cost == 50394

    d.read_io_per_s = 50000
    d.write_io_per_s = 34000
    assert d.annual_cost == 61110

    d.read_io_per_s = 75000
    d.write_io_per_s = 34000
    # (32000 * 0.78 + (64000-32000) * 0.552 + (75000 - 64000) * 0.384)
    # + ((32000 * 0.78) + (34000 - 32000) * 0.552)
    # 150
    assert d.annual_cost == 73062

    d.read_io_per_s = 256000
    d.write_io_per_s = 34000
    assert d.annual_cost == 142566



from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple

from .stateless_java import nflx_java_app_capacity_model
from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import FixedInterval
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.models import CapacityModel


class NflxWALCapacityModel(CapacityModel):
    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:

        wal_app = nflx_java_app_capacity_model.capacity_plan(
            instance=instance,
            drive=drive,
            context=context,
            desires=desires,
            extra_model_arguments=extra_model_arguments,
        )
        if wal_app is None:
            return None

        for cluster in wal_app.candidate_clusters.regional:
            cluster.cluster_type = "dgwwal"
        return wal_app

    @staticmethod
    def description():
        return "Netflix Streaming WAL Model"

    @staticmethod
    def extra_model_arguments_schema() -> Dict[str, Any]:
        return nflx_java_app_capacity_model.extra_model_arguments_schema()

    @staticmethod
    def compose_with(
        user_desires: CapacityDesires, extra_model_arguments: Dict[str, Any]
    ) -> Tuple[Tuple[str, Callable[[CapacityDesires], CapacityDesires]], ...]:
        provision_kv_shard = extra_model_arguments.get("wal.provisionkvshard", False)
        if provision_kv_shard:
            return (("org.netflix.key-value", lambda x: x),)
        else:
            return ()

    @staticmethod
    def default_desires(user_desires, extra_model_arguments):
        return CapacityDesires(
            query_pattern=QueryPattern(
                access_pattern=AccessPattern.latency,
                access_consistency=GlobalConsistency(
                    same_region=Consistency(
                        target_consistency=AccessConsistency.read_your_writes,
                    ),
                    cross_region=Consistency(
                        target_consistency=AccessConsistency.eventual,
                    ),
                ),
                estimated_mean_read_size_bytes=Interval(
                    low=128, mid=1024, high=65536, confidence=0.95
                ),
                estimated_mean_write_size_bytes=Interval(
                    low=64, mid=128, high=1024, confidence=0.95
                ),
                estimated_mean_read_latency_ms=Interval(
                    low=0.2, mid=4, high=6, confidence=0.98
                ),
                estimated_mean_write_latency_ms=Interval(
                    low=0.2, mid=1, high=2, confidence=0.98
                ),
                read_latency_slo_ms=FixedInterval(
                    minimum_value=1,
                    maximum_value=100,
                    low=1,
                    mid=8,
                    high=90,
                    confidence=0.98,
                ),
                write_latency_slo_ms=FixedInterval(
                    minimum_value=1,
                    maximum_value=20,
                    low=2,
                    mid=4,
                    high=10,
                    confidence=0.98,
                ),
            ),
            # Most throughput WAL clusters are large
            data_shape=DataShape(
                estimated_state_size_gib=Interval(
                    low=100, mid=1000, high=4000, confidence=0.98
                ),
                reserved_instance_app_mem_gib=8,
            ),
        )


nflx_wal_capacity_model = NflxWALCapacityModel()



from datetime import timedelta
from enum import Enum
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple

from pydantic import Field

from .stateless_java import nflx_java_app_capacity_model
from .stateless_java import NflxJavaAppArguments
from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import FixedInterval
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.models import CapacityModel


class NflxCounterCardinality(Enum):
    low = "low"
    medium = "medium"
    high = "high"


class NflxCounterMode(Enum):
    best_effort = "best-effort"
    eventual = "eventual"
    exact = "exact"


class NflxCounterArguments(NflxJavaAppArguments):
    counter_global: bool = Field(
        alias="counter.global",
        description="Indicate if this use case requires global counts which "
        "is more expensive than regional",
    )
    counter_cardinality: NflxCounterCardinality = Field(
        alias="counter.cardinality",
        description="Low means < 1,000, medium (1,000—100,000), high means > 100,000.",
    )
    counter_mode: NflxCounterMode = Field(
        alias="counter.mode",
        description="What mode of counting",
    )


class NflxCounterCapacityModel(CapacityModel):
    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:
        # Counter wants 20GiB root volumes
        extra_model_arguments.setdefault("root_disk_gib", 20)

        counter_app = nflx_java_app_capacity_model.capacity_plan(
            instance=instance,
            drive=drive,
            context=context,
            desires=desires,
            extra_model_arguments=extra_model_arguments,
        )
        if counter_app is None:
            return None

        for cluster in counter_app.candidate_clusters.regional:
            cluster.cluster_type = "dgwcounter"
        return counter_app

    @staticmethod
    def description():
        return "Netflix Streaming Counter Model"

    @staticmethod
    def extra_model_arguments_schema() -> Dict[str, Any]:
        return NflxCounterArguments.model_json_schema()

    @staticmethod
    def compose_with(
        user_desires: CapacityDesires, extra_model_arguments: Dict[str, Any]
    ) -> Tuple[Tuple[str, Callable[[CapacityDesires], CapacityDesires]], ...]:
        stores = [("org.netflix.evcache", lambda x: x)]
        if extra_model_arguments["counter.mode"] != NflxCounterMode.best_effort.value:

            def _modify_cassandra_desires(
                user_desires: CapacityDesires,
            ) -> CapacityDesires:
                modified = user_desires.model_copy(deep=True)

                # counts per second
                cps = user_desires.query_pattern.estimated_write_per_second

                # rollups happen once every 10 seconds after a write
                rps = cps.scale(0.1)
                modified.query_pattern.estimated_read_per_second = rps

                # storage size fix
                event_size = 128  # bytes
                count_size = 64  # bytes
                GiB = 1024 * 1024 * 1024
                retention = timedelta(days=1).total_seconds()
                cardinality = {
                    "low": 1_000,
                    "medium": 10_000,
                    "high": 100_000,
                }[extra_model_arguments["counter.cardinality"]]

                event_store = cps.scale(count_size * retention / GiB)
                count_store = event_size * cardinality / GiB
                store = event_store.offset(count_store)
                modified.data_shape.estimated_state_size_gib = store

                return modified

            stores.append(("org.netflix.cassandra", _modify_cassandra_desires))
        return tuple(stores)

    @staticmethod
    def default_desires(user_desires, extra_model_arguments):
        if user_desires.query_pattern.access_pattern == AccessPattern.latency:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.latency,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=64, mid=128, high=1024, confidence=0.95
                    ),
                    estimated_mean_read_latency_ms=Interval(
                        low=0.2, mid=1, high=2, confidence=0.98
                    ),
                    estimated_mean_write_latency_ms=Interval(
                        low=0.2, mid=1, high=2, confidence=0.98
                    ),
                    # "Single digit milliseconds SLO"
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=0.2,
                        maximum_value=10,
                        low=1,
                        mid=3,
                        high=6,
                        confidence=0.98,
                    ),
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=0.2,
                        maximum_value=10,
                        low=0.4,
                        mid=2,
                        high=5,
                        confidence=0.98,
                    ),
                ),
                data_shape=DataShape(
                    estimated_state_size_gib=Interval(
                        low=10, mid=50, high=200, confidence=0.98
                    ),
                    reserved_instance_app_mem_gib=8,
                ),
            )
        else:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.latency,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=64, mid=128, high=1024, confidence=0.95
                    ),
                    # counter scan queries can be more expensive
                    estimated_mean_read_latency_ms=Interval(
                        low=0.2, mid=4, high=6, confidence=0.98
                    ),
                    estimated_mean_write_latency_ms=Interval(
                        low=0.2, mid=1, high=2, confidence=0.98
                    ),
                    # Assume they're doing GetItems scans -> slow reads
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=1,
                        maximum_value=100,
                        low=1,
                        mid=8,
                        high=90,
                        confidence=0.98,
                    ),
                    # Assume they're doing PutRecords (BATCH)
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=1,
                        maximum_value=20,
                        low=2,
                        mid=4,
                        high=10,
                        confidence=0.98,
                    ),
                ),
                # Most throughput counter clusters are large
                data_shape=DataShape(
                    estimated_state_size_gib=Interval(
                        low=100, mid=1000, high=4000, confidence=0.98
                    ),
                    reserved_instance_app_mem_gib=8,
                ),
            )


nflx_counter_capacity_model = NflxCounterCapacityModel()



from decimal import Decimal

from service_capacity_modeling.hardware import shapes
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import CapacityRequirement
from service_capacity_modeling.interface import Clusters
from service_capacity_modeling.interface import default_reference_shape
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionClusterCapacity
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.interface import Requirements
from service_capacity_modeling.interface import ZoneClusterCapacity
from service_capacity_modeling.models.common import merge_plan
from service_capacity_modeling.models.common import network_services
from service_capacity_modeling.models.common import normalize_cores
from service_capacity_modeling.models.common import sqrt_staffed_cores


def test_merge_plan():
    left_requirement = CapacityRequirement(
        requirement_type="test",
        reference_shape=default_reference_shape,
        cpu_cores=Interval(low=10, mid=20, high=30, confidence=0.98),
        mem_gib=Interval(low=20, mid=100, high=200, confidence=0.98),
        network_mbps=Interval(low=1000, mid=2000, high=3000, confidence=0.98),
        disk_gib=Interval(low=40, mid=200, high=500, confidence=0.98),
    )
    right_requirement = CapacityRequirement(
        requirement_type="test",
        reference_shape=default_reference_shape,
        cpu_cores=Interval(low=10, mid=20, high=30, confidence=0.98),
        mem_gib=Interval(low=20, mid=100, high=200, confidence=0.98),
        network_mbps=Interval(low=1000, mid=2000, high=3000, confidence=0.98),
        disk_gib=Interval(low=40, mid=200, high=500, confidence=0.98),
    )

    left_instance = shapes.region("us-east-1").instances["r5d.2xlarge"]
    right_instance = shapes.region("us-east-1").instances["m5.2xlarge"]

    left_plan = CapacityPlan(
        requirements=Requirements(zonal=[left_requirement]),
        candidate_clusters=Clusters(
            annual_costs={"left-zonal": Decimal(1234)},
            zonal=[
                ZoneClusterCapacity(
                    cluster_type="left",
                    count=2,
                    instance=left_instance,
                    attached_drives=[],
                    annual_cost=1234,
                )
            ],
        ),
    )

    right_plan = CapacityPlan(
        requirements=Requirements(zonal=[right_requirement]),
        candidate_clusters=Clusters(
            annual_costs={"right-zonal": Decimal(1234), "right-regional": Decimal(234)},
            regional=[
                RegionClusterCapacity(
                    cluster_type="right",
                    count=2,
                    instance=right_instance,
                    attached_drives=[],
                    annual_cost=234,
                )
            ],
            zonal=[
                ZoneClusterCapacity(
                    cluster_type="right",
                    count=4,
                    instance=left_instance,
                    attached_drives=[],
                    annual_cost=1234,
                )
            ],
        ),
    )

    result = merge_plan(left_plan, right_plan)
    assert result is not None

    assert (
        result.requirements.zonal[0].cpu_cores.mid
        + result.requirements.zonal[1].cpu_cores.mid
        == 40
    )
    assert (
        result.requirements.zonal[0].network_mbps.mid
        + result.requirements.zonal[1].network_mbps.mid
        == 4000
    )

    assert result.candidate_clusters.regional == right_plan.candidate_clusters.regional

    assert left_plan.candidate_clusters.zonal[0] in result.candidate_clusters.zonal
    assert right_plan.candidate_clusters.zonal[0] in result.candidate_clusters.zonal
    assert "left-zonal" in result.candidate_clusters.annual_costs
    assert "right-zonal" in result.candidate_clusters.annual_costs
    assert result.candidate_clusters.total_annual_cost == (
        Decimal(1468) + Decimal(1234)
    )


def test_interval_scale():
    without_minmax = Interval(low=10, mid=100, high=1000)
    s = without_minmax.scale(3)
    assert s.low == 30
    assert s.mid == 300
    assert s.high == 3000
    assert s.minimum < s.low
    assert s.maximum > s.high

    with_minmax = Interval(
        low=10, mid=100, high=1000, minimum_value=1, maximum_value=1010
    )
    s = with_minmax.scale(3)
    assert s.low == 30
    assert s.mid == 300
    assert s.high == 3000
    assert s.minimum == 3
    assert s.maximum == 3030


def test_interval_offset():
    without_minmax = Interval(low=10, mid=100, high=1000)
    s = without_minmax.offset(3)
    assert s.low == 13
    assert s.mid == 103
    assert s.high == 1003
    assert s.minimum < s.low
    assert s.maximum > s.high

    with_minmax = Interval(
        low=10, mid=100, high=1000, minimum_value=1, maximum_value=1010
    )
    s = with_minmax.offset(3)
    assert s.low == 13
    assert s.mid == 103
    assert s.high == 1003
    assert s.minimum == 4
    assert s.maximum == 1013


def test_network_services():
    desires = CapacityDesires(
        service_tier=1,
        query_pattern=QueryPattern(
            estimated_write_per_second=Interval(
                low=1000, mid=10000, high=100000, confidence=0.98
            ),
            estimated_mean_write_size_bytes=Interval(
                low=128, mid=256, high=1024, confidence=0.98
            ),
        ),
    )
    hardware = shapes.region("us-east-1")
    region_context = RegionContext(
        services={n: s.model_copy(deep=True) for n, s in hardware.services.items()},
        num_regions=4,
        zones_in_region=3,
    )
    ns = network_services(
        "test", context=region_context, desires=desires, copies_per_region=3
    )
    cost_by_service = {}
    for service in ns:
        cost_by_service[service.service_type] = service.annual_cost

    assert 3 * 1500 < cost_by_service["test.net.inter.region"] < 3 * 1500 + 100
    assert 2 * 4 * 1500 < cost_by_service["test.net.intra.region"] < 2 * 4 * 1500 + 100


def test_different_tier_qos():
    tiers = (3, 2, 1, 0)
    prev_cores = 0
    for tier in tiers:
        desires = CapacityDesires(
            service_tier=tier,
            query_pattern=QueryPattern(
                estimated_read_per_second=Interval(
                    low=1000, mid=10000, high=100000, confidence=0.98
                ),
                estimated_write_per_second=Interval(
                    low=1000, mid=10000, high=100000, confidence=0.98
                ),
            ),
        )
        cores = sqrt_staffed_cores(desires)
        assert cores >= prev_cores
        prev_cores = cores


def test_normalize_cores():
    m5xl = shapes.region("us-east-1").instances["m5.xlarge"]
    r5xl = shapes.region("us-east-1").instances["r5.xlarge"]

    m6id = shapes.region("us-east-1").instances["m6id.xlarge"]
    i4ixl = shapes.region("us-east-1").instances["i4i.xlarge"]

    m7axl = shapes.region("us-east-1").instances["m7a.xlarge"]

    # Same generation should be the same
    assert normalize_cores(16, m5xl, r5xl) == 16
    assert normalize_cores(16, m6id, i4ixl) == 16

    # New generation should be higher
    assert 18 <= normalize_cores(16, m5xl, m6id) < 20
    assert 23 <= normalize_cores(16, m6id, m7axl) < 27
    assert 26 <= normalize_cores(16, m5xl, m7axl) < 30

    # All of these computers are much faster than the reference
    for shape in (m5xl, r5xl, m6id, i4ixl):
        assert normalize_cores(16, shape, default_reference_shape) < 13


def test_normalize_cores_6_7():
    """Note this is rather fragile to the exact math of normalize_cores

    If that method changes we'll need to change these assertions
    """
    m6ixl = shapes.region("us-east-1").instances["m6i.xlarge"]
    m7axl = shapes.region("us-east-1").instances["m7a.xlarge"]

    assert 11 == normalize_cores(16, m6ixl, default_reference_shape)
    assert 7 == normalize_cores(16, m7axl, default_reference_shape)



from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple

from .stateless_java import nflx_java_app_capacity_model
from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import FixedInterval
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.models import CapacityModel


class NflxKeyValueCapacityModel(CapacityModel):
    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:
        # KeyValue wants 20GiB root volumes
        extra_model_arguments.setdefault("root_disk_gib", 20)

        kv_app = nflx_java_app_capacity_model.capacity_plan(
            instance=instance,
            drive=drive,
            context=context,
            desires=desires,
            extra_model_arguments=extra_model_arguments,
        )
        if kv_app is None:
            return None

        for cluster in kv_app.candidate_clusters.regional:
            cluster.cluster_type = "dgwkv"
        return kv_app

    @staticmethod
    def description():
        return "Netflix Streaming Key-Value Model"

    @staticmethod
    def extra_model_arguments_schema() -> Dict[str, Any]:
        return nflx_java_app_capacity_model.extra_model_arguments_schema()

    @staticmethod
    def compose_with(
        user_desires: CapacityDesires, extra_model_arguments: Dict[str, Any]
    ) -> Tuple[Tuple[str, Callable[[CapacityDesires], CapacityDesires]], ...]:
        query_pattern = user_desires.query_pattern
        target_consistency = (
            query_pattern.access_consistency.same_region.target_consistency
        )
        rps_interval = query_pattern.estimated_read_per_second
        rps: float = rps_interval.mid
        wps: float = query_pattern.estimated_write_per_second.mid
        read_write_ratio: float = rps / wps

        # Parameterizing this in case we want to configure it to something else later.
        # The read/write ratio should be relatively high to make EVCache effective.
        evcache_rw_ratio_threshold: float = extra_model_arguments.get(
            "kv_evcache_read_write_ratio_threshold", 0.9
        )
        use_evcache = target_consistency in (
            AccessConsistency.eventual,
            AccessConsistency.best_effort,
        ) and (
            rps > 250_000
            or (rps > 100_000 and read_write_ratio > evcache_rw_ratio_threshold)
        )

        if use_evcache:

            def _modify_cassandra_desires(
                desires: CapacityDesires,
            ) -> CapacityDesires:
                relaxed = desires.model_copy(deep=True)

                # This is an initial best guess. Parameterizing in case we want to
                # configure it in the future.
                estimated_kv_cache_hit_rate: float = extra_model_arguments.get(
                    "estimated_kv_cache_hit_rate", 0.8
                )

                # Scale down the Cassandra estimated rps since those reads will be
                # serviced by EVCache.
                relaxed.query_pattern.estimated_read_per_second = rps_interval.scale(
                    1 - estimated_kv_cache_hit_rate
                )
                return relaxed

            def _modify_evcache_desires(
                desires: CapacityDesires,
            ) -> CapacityDesires:
                relaxed = desires.model_copy(deep=True)
                access_consistency = relaxed.query_pattern.access_consistency
                access_consistency.same_region.target_consistency = (
                    AccessConsistency.best_effort
                )
                return relaxed

            return (
                ("org.netflix.cassandra", _modify_cassandra_desires),
                ("org.netflix.evcache", _modify_evcache_desires),
            )
        else:
            return (("org.netflix.cassandra", lambda x: x),)

    @staticmethod
    def default_desires(user_desires, extra_model_arguments):
        if user_desires.query_pattern.access_pattern == AccessPattern.latency:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.latency,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.read_your_writes,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=64, mid=128, high=1024, confidence=0.95
                    ),
                    estimated_mean_read_latency_ms=Interval(
                        low=0.2, mid=1, high=2, confidence=0.98
                    ),
                    estimated_mean_write_latency_ms=Interval(
                        low=0.2, mid=1, high=2, confidence=0.98
                    ),
                    # "Single digit milliseconds SLO"
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=0.2,
                        maximum_value=10,
                        low=1,
                        mid=3,
                        high=6,
                        confidence=0.98,
                    ),
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=0.2,
                        maximum_value=10,
                        low=0.4,
                        mid=2,
                        high=5,
                        confidence=0.98,
                    ),
                ),
                # Most KeyValue clusters are small
                data_shape=DataShape(
                    estimated_state_size_gib=Interval(
                        low=10, mid=50, high=200, confidence=0.98
                    ),
                    reserved_instance_app_mem_gib=8,
                ),
            )
        else:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.latency,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.read_your_writes,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=64, mid=128, high=1024, confidence=0.95
                    ),
                    # KV scan queries can be more expensive
                    estimated_mean_read_latency_ms=Interval(
                        low=0.2, mid=4, high=6, confidence=0.98
                    ),
                    estimated_mean_write_latency_ms=Interval(
                        low=0.2, mid=1, high=2, confidence=0.98
                    ),
                    # Assume they're doing GetItems scans -> slow reads
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=1,
                        maximum_value=100,
                        low=1,
                        mid=8,
                        high=90,
                        confidence=0.98,
                    ),
                    # Assume they're doing PutRecords (BATCH)
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=1,
                        maximum_value=20,
                        low=2,
                        mid=4,
                        high=10,
                        confidence=0.98,
                    ),
                ),
                # Most throughput KV clusters are large
                data_shape=DataShape(
                    estimated_state_size_gib=Interval(
                        low=100, mid=1000, high=4000, confidence=0.98
                    ),
                    reserved_instance_app_mem_gib=8,
                ),
            )


nflx_key_value_capacity_model = NflxKeyValueCapacityModel()



# pylint: disable=cyclic-import
# in HardwareShapes.hardware it imports from hardware.profiles dynamically
import json
import logging
import os
from functools import reduce
from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import GlobalHardware
from service_capacity_modeling.interface import Hardware
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Pricing
from service_capacity_modeling.interface import Service

logger = logging.getLogger(__name__)


def load_pricing(pricing: Dict) -> Pricing:
    return Pricing(regions=pricing)


def load_hardware(hardware: Dict) -> Hardware:
    return Hardware(**hardware)


def price_hardware(hardware: Hardware, pricing: Pricing) -> GlobalHardware:
    regions: Dict[str, Hardware] = {}

    for region, region_pricing in pricing.regions.items():
        priced_instances: Dict[str, Instance] = {}
        priced_drives: Dict[str, Drive] = {}
        priced_services: Dict[str, Service] = {}

        for instance, iprice in region_pricing.instances.items():
            if instance not in hardware.instances:
                logger.debug(
                    "Instance %s is not in hardware shapes, skipping", instance
                )
                continue
            priced_instances[instance] = hardware.instances[instance].model_copy()
            priced_instances[instance].annual_cost = iprice.annual_cost
            if iprice.lifecycle is not None:
                priced_instances[instance].lifecycle = iprice.lifecycle

        for drive, dprice in region_pricing.drives.items():
            if drive not in hardware.drives:
                logger.warning("Drive %s is not in hardware shapes, skipping", drive)
                continue
            priced_drives[drive] = hardware.drives[drive].model_copy()
            priced_drives[drive].annual_cost_per_gib = dprice.annual_cost_per_gib
            priced_drives[
                drive
            ].annual_cost_per_read_io = dprice.annual_cost_per_read_io
            priced_drives[
                drive
            ].annual_cost_per_write_io = dprice.annual_cost_per_write_io

        for svc, svc_price in region_pricing.services.items():
            if svc not in hardware.services:
                logger.warning("Service %s is not in hardware shapes, skipping", svc)
                continue
            priced_services[svc] = hardware.services[svc].model_copy()
            priced_services[svc].annual_cost_per_gib = svc_price.annual_cost_per_gib
            priced_services[
                svc
            ].annual_cost_per_read_io = svc_price.annual_cost_per_read_io
            priced_services[
                svc
            ].annual_cost_per_write_io = svc_price.annual_cost_per_write_io
            priced_services[svc].annual_cost_per_core = svc_price.annual_cost_per_core

        regions[region] = Hardware(
            instances=priced_instances,
            drives=priced_drives,
            services=priced_services,
            zones_in_region=region_pricing.zones_in_region,
        )

    return GlobalHardware(regions=regions)


def merge_hardware(existing: Hardware, override: Hardware) -> Hardware:
    """Merge two hardware sets, Unlike pricing does _not_ support overrides

    This method takes two shape files and merges instances, drives, services, etc.
    Unlike pricing this does not override - it will throw exceptions if two files
    contain the same keys
    """
    existing_obj = existing.model_dump()
    override_obj = override.model_dump()

    merged = {}
    for key in existing_obj.keys() | override_obj.keys():
        existing_field = existing_obj.get(key)
        if existing_field is None and override_obj.get(key) is None:
            continue
        if existing_field is None and override_obj.get(key) is not None:
            merged[key] = override_obj.get(key)
        elif isinstance(existing_field, Dict):
            override_field = override_obj.get(key)
            merged_field = merged.setdefault(key, {})

            existing_keys = existing_field.keys()
            override_keys = override_obj.get(key, {}).keys()
            for shape in existing_keys | override_keys:
                if shape in existing_keys and shape in override_keys:
                    raise ValueError(
                        f"Duplicate shape {shape}! "
                        "Only one file should contain a shape"
                    )
                if shape not in existing_keys:
                    merged_field[shape] = override_field[shape]
                else:
                    merged_field[shape] = existing_field[shape]
    return Hardware(**merged)


def merge_pricing(existing: Dict, override: Dict) -> Dict:
    merged = existing.copy()
    for region, override_pricing in override.items():
        if region not in merged:
            merged[region] = override_pricing
        else:
            for instance, instance_data in override_pricing["instances"].items():
                if merged[region]["instances"].get(instance) is None:
                    merged[region]["instances"][instance] = instance_data.copy()
                else:
                    merged[region]["instances"][instance].update(instance_data)

            for drive, dprice in override_pricing["drives"].items():
                if merged[region]["drives"].get(drive) is None:
                    merged[region]["drives"][drive] = dprice.copy()
                else:
                    merged[region]["drives"][drive].update(dprice)

            for service, sprice in override_pricing["services"].items():
                if merged[region]["services"].get(service) is None:
                    merged[region]["services"][service] = sprice.copy()
                else:
                    merged[region]["services"][service].update(sprice)

    return merged


def load_hardware_from_disk(
    price_paths: Union[List[Path], Optional[str]] = os.environ.get("PRICE_PATH"),
    shape_paths: Union[List[Path], Optional[str]] = os.environ.get("HARDWARE_SHAPES"),
) -> GlobalHardware:
    if isinstance(shape_paths, list) and len(shape_paths) == 0:
        return GlobalHardware(regions={})
    if isinstance(price_paths, list) and len(price_paths) == 0:
        return GlobalHardware(regions={})

    if isinstance(shape_paths, str):
        shape_paths = [Path(shape_paths)]
    if isinstance(price_paths, str):
        price_paths = [Path(price_paths)]

    if price_paths is None:
        price_paths = []
    if shape_paths is None:
        shape_paths = []

    combined_pricing: Dict = {}

    logger.debug("Loading pricing from: %s", price_paths)
    for price_path in price_paths:
        logger.debug("Loading pricing from: %s", price_path)
        with open(price_path, encoding="utf-8") as pfd:
            pricing_data = json.load(pfd)
            combined_pricing = merge_pricing(combined_pricing, pricing_data)

    pricing = load_pricing(combined_pricing)

    hw_shapes = [Hardware()]
    for path in shape_paths:
        with open(path, encoding="utf-8") as fd:
            hw_shapes.append(Hardware(**json.load(fd)))

    hardware = reduce(merge_hardware, hw_shapes)
    return price_hardware(hardware=hardware, pricing=pricing)


def load_hardware_from_s3(bucket, path) -> GlobalHardware:
    try:
        # boto is a heavy dependency so we only want to take it if
        # someone will be using it ...
        try:
            import boto3
            import botocore
        except ImportError:
            return GlobalHardware(regions={})

        s3 = boto3.resource("s3")
        obj = s3.Object(bucket, path)
        data = json.loads(obj.get()["Body"].read().decode("utf-8"))
        return GlobalHardware(**data)
    except botocore.exceptions.ClientError as exp:
        logger.exception(exp)
        return GlobalHardware(regions={})


class HardwareShapes:
    def __init__(self):
        self._hardware: Optional[GlobalHardware] = None

    def load(self, new_hardware: GlobalHardware) -> None:
        self._hardware = new_hardware

    @property
    def hardware(self) -> GlobalHardware:
        if self._hardware is None:
            from service_capacity_modeling.hardware.profiles import common_profiles

            self._hardware = common_profiles["aws-3yr-reserved"]
        return self._hardware

    def region(self, region: str) -> Hardware:
        return self.hardware.regions[region]

    def instance(self, name: str, region: Optional[str] = None) -> Instance:
        if region is not None:
            return self.region(region).instances[name]

        for _, hw in self.hardware.regions.items():
            if name in hw.instances:
                return hw.instances[name]

        raise KeyError(f"Unknown instance {name}")


shapes: HardwareShapes = HardwareShapes()



import logging
import math
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional

from pydantic import BaseModel
from pydantic import Field

from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import Buffer
from service_capacity_modeling.interface import BufferComponent
from service_capacity_modeling.interface import Buffers
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import CapacityRequirement
from service_capacity_modeling.interface import certain_float
from service_capacity_modeling.interface import certain_int
from service_capacity_modeling.interface import Clusters
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import FixedInterval
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.interface import Requirements
from service_capacity_modeling.interface import ServiceCapacity
from service_capacity_modeling.models import CapacityModel
from service_capacity_modeling.models.common import buffer_for_components
from service_capacity_modeling.models.common import compute_stateful_zone
from service_capacity_modeling.models.common import network_services
from service_capacity_modeling.models.common import normalize_cores
from service_capacity_modeling.models.common import simple_network_mbps
from service_capacity_modeling.models.common import sqrt_staffed_cores
from service_capacity_modeling.models.common import working_set_from_drive_and_slo
from service_capacity_modeling.models.utils import next_power_of_2
from service_capacity_modeling.stats import dist_for_interval

logger = logging.getLogger(__name__)

BACKGROUND_BUFFER = "background"


def _write_buffer_gib_zone(
    desires: CapacityDesires, zones_per_region: int, flushes_before_compaction: int = 4
) -> float:
    # Cassandra has to buffer writes before flushing to disk, and assuming
    # we will compact every 4 flushes and we want no more than 2 redundant
    # compactions in an hour, we want <= 4**2 = 16 flushes per hour
    # or a flush of data every 3600 / 16 = 225 seconds
    write_bytes_per_second = (
        desires.query_pattern.estimated_write_per_second.mid
        * desires.query_pattern.estimated_mean_write_size_bytes.mid
    )

    compactions_per_hour = 2
    hour_in_seconds = 60 * 60

    write_buffer_gib = (
        (write_bytes_per_second * hour_in_seconds)
        / (flushes_before_compaction**compactions_per_hour)
    ) / (1 << 30)

    return float(write_buffer_gib) / zones_per_region


def _estimate_cassandra_requirement(  # pylint: disable=too-many-positional-arguments
    instance: Instance,
    desires: CapacityDesires,
    working_set: float,
    reads_per_second: float,
    max_rps_to_disk: int,
    required_cluster_size: Optional[int] = None,
    zones_per_region: int = 3,
    copies_per_region: int = 3,
) -> CapacityRequirement:
    """Estimate the capacity required for one zone given a regional desire

    The input desires should be the **regional** desire, and this function will
    return the zonal capacity requirement
    """
    current_capacity = (
        None
        if desires.current_clusters is None
        else (
            desires.current_clusters.zonal[0]
            if len(desires.current_clusters.zonal)
            else desires.current_clusters.regional[0]
        )
    )
    # Cassandra does various background activity (compaction, repair, backup) which
    # require up to half a machine's cores and network to safely perform.
    background_buffer = buffer_for_components(
        buffers=desires.buffers, components=[BACKGROUND_BUFFER]
    )

    # Currently, zones and regions are configured in a homogeneous manner. Hence,
    # we just take any one of the current cluster configuration
    if (
        current_capacity
        and current_capacity.cluster_instance
        and required_cluster_size is not None
    ):
        needed_cores = int(
            (
                current_capacity.cluster_instance.cpu
                * required_cluster_size
                * zones_per_region
            )
            * (current_capacity.cpu_utilization.high / 20)
        )
        # Normalize those cores to the target shape
        reference_shape = desires.reference_shape
        needed_cores = normalize_cores(
            core_count=needed_cores,
            target_shape=instance,
            reference_shape=reference_shape,
        )
        reference_shape = current_capacity.cluster_instance
    else:
        # We have no existing utilization to go from
        reference_shape = desires.reference_shape
        # Keep half of the cores free for background work (compaction, backup, repair).
        needed_cores = math.ceil(sqrt_staffed_cores(desires) * background_buffer.ratio)

        needed_cores = normalize_cores(
            core_count=needed_cores,
            target_shape=instance,
            reference_shape=reference_shape,
        )

    # Keep half of the bandwidth available for backup and repair streaming
    needed_network_mbps = simple_network_mbps(desires) * background_buffer.ratio

    needed_disk = math.ceil(
        (1.0 / desires.data_shape.estimated_compression_ratio.mid)
        * desires.data_shape.estimated_state_size_gib.mid
        * copies_per_region,
    )

    # Rough estimate of how many instances we would need just for the CPU
    # Note that this is a lower bound, we might end up with more.
    rough_count = math.ceil(needed_cores / instance.cpu)

    # Generally speaking we want fewer than some number of reads per second
    # hitting disk per instance. If we don't have many reads we don't need to
    # hold much data in memory.
    instance_rps = max(1, reads_per_second // rough_count)
    disk_rps = instance_rps * _cass_io_per_read(max(1, needed_disk // rough_count))
    rps_working_set = min(1.0, disk_rps / max_rps_to_disk)

    # If disk RPS will be smaller than our target because there are no
    # reads, we don't need to hold as much data in memory
    needed_memory = min(working_set, rps_working_set) * needed_disk

    # Now convert to per zone
    needed_cores = max(1, needed_cores // zones_per_region)
    needed_disk = max(1, needed_disk // zones_per_region)
    needed_memory = max(1, int(needed_memory // zones_per_region))
    logger.debug(
        "Need (cpu, mem, disk, working) = (%s, %s, %s, %f)",
        needed_cores,
        needed_memory,
        needed_disk,
        working_set,
    )

    # Cassandra can defer writes either by buffering in memory or by
    # waiting longer before recompacting (the min-threshold on the
    # L0 compactions or STCS compactions)
    min_threshold = 4
    write_buffer_gib = _write_buffer_gib_zone(
        desires=desires,
        zones_per_region=zones_per_region,
        flushes_before_compaction=min_threshold,
    )

    while write_buffer_gib > 12 and min_threshold < 16:
        min_threshold *= 2
        write_buffer_gib = _write_buffer_gib_zone(
            desires=desires,
            zones_per_region=zones_per_region,
            flushes_before_compaction=min_threshold,
        )

    return CapacityRequirement(
        requirement_type="cassandra-zonal",
        reference_shape=reference_shape,
        cpu_cores=certain_int(needed_cores),
        mem_gib=certain_float(needed_memory),
        disk_gib=certain_float(needed_disk),
        network_mbps=certain_float(needed_network_mbps),
        context={
            "working_set": min(working_set, rps_working_set),
            "rps_working_set": rps_working_set,
            "disk_slo_working_set": working_set,
            "replication_factor": copies_per_region,
            "compression_ratio": round(
                1.0 / desires.data_shape.estimated_compression_ratio.mid, 2
            ),
            "read_per_second": reads_per_second,
            "write_buffer_gib": write_buffer_gib,
            "min_threshold": min_threshold,
        },
    )


def _upsert_params(cluster, params):
    if cluster.cluster_params:
        cluster.cluster_params.update(params)
    else:
        cluster.cluster_params = params


# pylint: disable=too-many-locals
# pylint: disable=too-many-return-statements
# flake8: noqa: C901
def _estimate_cassandra_cluster_zonal(  # pylint: disable=too-many-positional-arguments
    instance: Instance,
    drive: Drive,
    context: RegionContext,
    desires: CapacityDesires,
    zones_per_region: int = 3,
    copies_per_region: int = 3,
    require_local_disks: bool = True,
    require_attached_disks: bool = False,
    required_cluster_size: Optional[int] = None,
    max_rps_to_disk: int = 500,
    max_local_disk_gib: int = 5120,
    max_regional_size: int = 192,
    max_write_buffer_percent: float = 0.25,
    max_table_buffer_percent: float = 0.11,
) -> Optional[CapacityPlan]:
    # Netflix Cassandra doesn't like to deploy on really small instances
    if instance.cpu < 2 or instance.ram_gib < 14:
        return None

    # if we're not allowed to use gp2, skip EBS only types
    if instance.drive is None and require_local_disks:
        return None

    # if we're not allowed to use local disks, skip ephems
    if instance.drive is not None and require_attached_disks:
        return None

    # Cassandra only deploys on gp2 and gp3 drives right now
    if drive.name not in ("gp2", "gp3"):
        return None

    rps = desires.query_pattern.estimated_read_per_second.mid // zones_per_region
    write_per_sec = (
        desires.query_pattern.estimated_write_per_second.mid // zones_per_region
    )
    write_bytes_per_sec = round(
        write_per_sec * desires.query_pattern.estimated_mean_write_size_bytes.mid
    )
    read_bytes_per_sec = rps * desires.query_pattern.estimated_mean_read_size_bytes.mid
    # Write IO will be 1 to commitlog + 2 writes (plus 2 reads) in the first
    # hour during compaction.
    # Writes are sequential
    # Reads are random
    write_io_per_sec = (1 + 4) * max(
        1, write_bytes_per_sec // (drive.seq_io_size_kib * 1024)
    )
    read_io_per_sec = max(rps, read_bytes_per_sec // (drive.rand_io_size_kib * 1024))

    # Based on the disk latency and the read latency SLOs we adjust our
    # working set to keep more or less data in RAM. Faster drives need
    # less fronting RAM.
    ws_drive = instance.drive or drive
    working_set = working_set_from_drive_and_slo(
        drive_read_latency_dist=dist_for_interval(ws_drive.read_io_latency_ms),
        read_slo_latency_dist=dist_for_interval(
            desires.query_pattern.read_latency_slo_ms
        ),
        estimated_working_set=desires.data_shape.estimated_working_set_percent,
        # This is about right for a database, a cache probably would want
        # to increase this even more.
        target_percentile=0.95,
    ).mid

    requirement = _estimate_cassandra_requirement(
        instance=instance,
        desires=desires,
        working_set=working_set,
        reads_per_second=rps,
        max_rps_to_disk=max_rps_to_disk,
        required_cluster_size=required_cluster_size,
        zones_per_region=zones_per_region,
        copies_per_region=copies_per_region,
    )

    # Cassandra clusters should aim to be at least 2 nodes per zone to start
    # out with for tier 0 or tier 1. This gives us more room to "up-color"]
    # clusters.
    min_count = 0
    if desires.service_tier <= 1:
        min_count = 2

    base_mem = (
        desires.data_shape.reserved_instance_app_mem_gib
        + desires.data_shape.reserved_instance_system_mem_gib
    )

    heap_fn = _cass_heap_for_write_buffer(
        instance=instance,
        max_zonal_size=max_regional_size // zones_per_region,
        write_buffer_gib=requirement.context["write_buffer_gib"],
        buffer_percent=(max_write_buffer_percent * max_table_buffer_percent),
    )

    # Typically 4x buffer except when rightsizing
    disk_buffer = buffer_for_components(
        buffers=desires.buffers, components=[BufferComponent.disk]
    )

    cluster = compute_stateful_zone(
        instance=instance,
        drive=drive,
        needed_cores=int(requirement.cpu_cores.mid),
        needed_disk_gib=int(requirement.disk_gib.mid),
        needed_memory_gib=int(requirement.mem_gib.mid),
        needed_network_mbps=requirement.network_mbps.mid,
        # Take into account the reads per read
        # from the per node dataset using leveled compaction
        required_disk_ios=lambda size, count: (
            _cass_io_per_read(size) * math.ceil(read_io_per_sec / count),
            write_io_per_sec / count,
        ),
        # C* requires ephemeral disks to be 25% full because compaction
        # and replacement time if we're underscaled.
        required_disk_space=lambda x: x * disk_buffer.ratio,
        # C* clusters cannot recover data from neighbors quickly so we
        # want to avoid clusters with more than 1 TiB of local state
        max_local_disk_gib=max_local_disk_gib,
        # C* clusters provision in powers of 2 because doubling
        cluster_size=next_power_of_2,
        min_count=max(min_count, required_cluster_size or 0),
        # C* heap usage takes away from OS page cache memory
        reserve_memory=lambda x: base_mem + heap_fn(x),
        # C* heap buffers the writes at roughly a rate of
        # memtable_cleanup_threshold * memtable_size. At Netflix this
        # is 0.11 * 25 * heap
        write_buffer=lambda x: heap_fn(x) * max_write_buffer_percent * 0.25,
        required_write_buffer_gib=float(requirement.context["write_buffer_gib"]),
    )

    # Communicate to the actual provision that if we want reduced RF
    params = {
        "cassandra.keyspace.rf": copies_per_region,
        # In order to handle high write loads we have to shift memory
        # to heap memory, communicate with C* about this
        "cassandra.heap.gib": heap_fn(instance.ram_gib),
        "cassandra.heap.write.percent": max_write_buffer_percent,
        "cassandra.heap.table.percent": max_table_buffer_percent,
        "cassandra.compaction.min_threshold": requirement.context["min_threshold"],
    }
    _upsert_params(cluster, params)

    # Sometimes we don't want modify cluster topology, so only allow
    # topologies that match the desired zone size
    if required_cluster_size is not None and cluster.count != required_cluster_size:
        return None

    # Cassandra clusters generally should try to stay under some total number
    # of nodes. Orgs do this for all kinds of reasons such as
    #   * Security group limits. Since you must have < 500 rules if you're
    #       ingressing public ips)
    #   * Maintenance. If your restart script does one node at a time you want
    #       smaller clusters so your restarts don't take months.
    #   * Schema propagation. Since C* must gossip out changes to schema the
    #       duration of this can increase a lot with > 500 node clusters.
    if cluster.count > (max_regional_size // zones_per_region):
        return None

    # Durable Cassandra clusters backup to S3
    # TODO use the write rate and estimated write size to estimate churn
    # over the retention period.
    cap_services = []
    if desires.data_shape.durability_slo_order.mid >= 1000:
        blob = context.services.get("blob.standard", None)
        if blob:
            cap_services = [
                ServiceCapacity(
                    service_type=f"cassandra.backup.{blob.name}",
                    annual_cost=blob.annual_cost_gib(requirement.disk_gib.mid),
                    service_params={
                        "nines_required": (
                            1 - 1.0 / desires.data_shape.durability_slo_order.mid
                        )
                    },
                )
            ]

    network_costs = network_services("cassandra", context, desires, copies_per_region)
    if network_costs:
        cap_services.extend(network_costs)

    # Account for the clusters, backup, and network costs
    cassandra_costs = {
        "cassandra.zonal-clusters": zones_per_region * cluster.annual_cost,
    }
    for s in cap_services:
        cassandra_costs[f"{s.service_type}"] = s.annual_cost

    cluster.cluster_type = "cassandra"
    clusters = Clusters(
        annual_costs=cassandra_costs,
        zonal=[cluster] * zones_per_region,
        regional=[],
        services=cap_services,
    )

    return CapacityPlan(
        requirements=Requirements(zonal=[requirement] * zones_per_region),
        candidate_clusters=clusters,
    )


# C* LCS has 160 MiB sstables by default and 10 sstables per level
def _cass_io_per_read(node_size_gib, sstable_size_mb=160):
    gb = node_size_gib * 1024
    sstables = max(1, gb // sstable_size_mb)
    # 10 sstables per level, plus 1 for L0 (avg)
    levels = 1 + int(math.ceil(math.log(sstables, 10)))
    # One disk IO per data read and one per index read (assume we miss
    # the key cache)
    return 2 * levels


def _cass_heap_for_write_buffer(
    instance: Instance,
    write_buffer_gib: float,
    max_zonal_size: int,
    buffer_percent: float,
) -> Callable[[float], float]:
    # If there is no way we can get enough heap with the max zonal size, try
    # letting max heap grow to 31 GiB per node to get more write buffer
    if write_buffer_gib > (
        max_zonal_size * _cass_heap(instance.ram_gib) * buffer_percent
    ):
        return lambda x: _cass_heap(x, max_heap_gib=30)
    else:
        return _cass_heap


# C* follows the following formula for calculating heap
def _cass_heap(node_memory_gib: float, max_heap_gib: float = 30) -> float:
    # OSS Cassandra does this
    # max(min(node_memory_gib // 2, 4), min(node_memory_gib // 4, max_heap_gib))

    # Netflix Cassandra does this
    return min(max(4, node_memory_gib // 2), max_heap_gib)


def _target_rf(desires: CapacityDesires, user_copies: Optional[int]) -> int:
    if user_copies is not None:
        assert user_copies > 1
        return user_copies

    # Due to the relaxed durability and consistency requirements we can
    # run with RF=2
    consistency = desires.query_pattern.access_consistency.same_region
    if (
        desires.data_shape.durability_slo_order.mid < 1000
        and consistency is not None
        and consistency.target_consistency != AccessConsistency.read_your_writes
    ):
        return 2
    return 3


class NflxCassandraArguments(BaseModel):
    copies_per_region: int = Field(
        default=3,
        description="How many copies of the data will exist e.g. RF=3. If unsupplied"
        " this will be deduced from durability and consistency desires",
    )
    require_local_disks: bool = Field(
        default=True,
        description="If local (ephemeral) drives are required",
    )
    require_attached_disks: bool = Field(
        default=False,
        description="If attached (ebs) drives are required",
    )
    required_cluster_size: Optional[int] = Field(
        default=None,
        description="Require zonal clusters to be this size (force vertical scaling)",
    )
    max_rps_to_disk: int = Field(
        default=500,
        description="How many disk IOs should be allowed to hit disk per instance",
    )
    max_regional_size: int = Field(
        default=192,
        description="What is the maximum size of a cluster in this region",
    )
    max_local_disk_gib: int = Field(
        default=5120,
        description="The maximum amount of data we store per machine",
    )
    max_write_buffer_percent: float = Field(
        default=0.25,
        description="The amount of heap memory that can be used to buffer writes. "
        "Note that if there are more than 100k writes this will "
        "automatically adjust to 0.5",
    )
    max_table_buffer_percent: float = Field(
        default=0.11,
        description="How much of heap memory can be used for a single table. "
        "Note that if there are more than 100k writes this will "
        "automatically adjust to 0.2",
    )


class NflxCassandraCapacityModel(CapacityModel):
    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:
        # Use durabiliy and consistency to compute RF.
        copies_per_region = _target_rf(
            desires, extra_model_arguments.get("copies_per_region", None)
        )
        require_local_disks: bool = extra_model_arguments.get(
            "require_local_disks", True
        )
        require_attached_disks: bool = extra_model_arguments.get(
            "require_attached_disks", False
        )
        required_cluster_size: Optional[int] = extra_model_arguments.get(
            "required_cluster_size", None
        )
        max_rps_to_disk: int = extra_model_arguments.get("max_rps_to_disk", 500)
        max_regional_size: int = extra_model_arguments.get("max_regional_size", 192)
        max_local_disk_gib: int = extra_model_arguments.get("max_local_disk_gib", 5120)
        max_write_buffer_percent: float = min(
            0.5, extra_model_arguments.get("max_write_buffer_percent", 0.25)
        )
        max_table_buffer_percent: float = min(
            0.5, extra_model_arguments.get("max_table_buffer_percent", 0.11)
        )

        # Adjust heap defaults for high write clusters
        if (
            desires.query_pattern.estimated_write_per_second.mid >= 100_000
            and desires.data_shape.estimated_state_size_gib.mid >= 100
        ):
            max_write_buffer_percent = max(0.5, max_write_buffer_percent)
            max_table_buffer_percent = max(0.2, max_table_buffer_percent)

        return _estimate_cassandra_cluster_zonal(
            instance=instance,
            drive=drive,
            context=context,
            desires=desires,
            zones_per_region=context.zones_in_region,
            copies_per_region=copies_per_region,
            require_local_disks=require_local_disks,
            require_attached_disks=require_attached_disks,
            required_cluster_size=required_cluster_size,
            max_rps_to_disk=max_rps_to_disk,
            max_regional_size=max_regional_size,
            max_local_disk_gib=max_local_disk_gib,
            max_write_buffer_percent=max_write_buffer_percent,
            max_table_buffer_percent=max_table_buffer_percent,
        )

    @staticmethod
    def description():
        return "Netflix Streaming Cassandra Model"

    @staticmethod
    def extra_model_arguments_schema() -> Dict[str, Any]:
        return NflxCassandraArguments.model_json_schema()

    @staticmethod
    def default_desires(user_desires, extra_model_arguments: Dict[str, Any]):
        acceptable_consistency = {
            None,
            AccessConsistency.best_effort,
            AccessConsistency.eventual,
            AccessConsistency.read_your_writes,
            AccessConsistency.never,
        }
        for key, value in user_desires.query_pattern.access_consistency:
            if value.target_consistency not in acceptable_consistency:
                raise ValueError(
                    f"Cassandra can only provide {acceptable_consistency} access."
                    f"User asked for {key}={value}"
                )

        # Lower RF = less write compute
        rf = _target_rf(
            user_desires, extra_model_arguments.get("copies_per_region", None)
        )
        if rf < 3:
            rf_write_latency = Interval(low=0.2, mid=0.6, high=2, confidence=0.98)
        else:
            rf_write_latency = Interval(low=0.4, mid=1, high=2, confidence=0.98)

        # By supplying these buffers we can deconstruct observed utilization into
        # load versus buffer.
        buffers = Buffers(
            default=Buffer(ratio=1.5),
            desired={
                "compute": Buffer(ratio=1.5, components=[BufferComponent.compute]),
                "storage": Buffer(ratio=4.0, components=[BufferComponent.storage]),
                # Cassandra reserves headroom in both cpu and network for background
                # work and tasks
                "background": Buffer(
                    ratio=2.0,
                    components=[
                        BufferComponent.cpu,
                        BufferComponent.network,
                        BACKGROUND_BUFFER,
                    ],
                ),
            },
        )

        if user_desires.query_pattern.access_pattern == AccessPattern.latency:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.latency,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.read_your_writes,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=64, mid=256, high=1024, confidence=0.95
                    ),
                    # Cassandra point queries usualy take just around 2ms
                    # of on CPU time for reads and 1ms for writes
                    estimated_mean_read_latency_ms=Interval(
                        low=0.4, mid=2, high=5, confidence=0.98
                    ),
                    estimated_mean_write_latency_ms=rf_write_latency,
                    # Assume point queries, "Single digit milliseconds SLO"
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=0.2,
                        maximum_value=10,
                        low=0.4,
                        mid=2,
                        high=5,
                        confidence=0.98,
                    ),
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=0.2,
                        maximum_value=10,
                        low=0.4,
                        mid=1,
                        high=4,
                        confidence=0.98,
                    ),
                ),
                # Most latency sensitive cassandra clusters are in the
                # < 1TiB range
                data_shape=DataShape(
                    estimated_state_size_gib=Interval(
                        low=10, mid=100, high=1000, confidence=0.98
                    ),
                    # Cassandra compresses with LZ4 by default
                    estimated_compression_ratio=Interval(
                        minimum_value=1.1,
                        maximum_value=8,
                        low=2,
                        mid=3,
                        high=5,
                        confidence=0.98,
                    ),
                    # We dynamically allocate the C* JVM memory in the plan
                    # but account for the Priam sidecar here
                    reserved_instance_app_mem_gib=4,
                ),
                buffers=buffers,
            )
        else:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.throughput,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.read_your_writes,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    # Cassandra scan queries usually take longer
                    estimated_mean_read_latency_ms=Interval(
                        low=0.2, mid=5, high=20, confidence=0.98
                    ),
                    # Usually throughput clusters are running RF=2
                    # Maybe revise this?
                    estimated_mean_write_latency_ms=Interval(
                        low=0.2, mid=0.6, high=2, confidence=0.98
                    ),
                    # Assume they're scanning -> slow reads
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=1,
                        maximum_value=100,
                        low=2,
                        mid=8,
                        high=90,
                        confidence=0.98,
                    ),
                    # Assume they're doing BATCH writes
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=0.5,
                        maximum_value=20,
                        low=1,
                        mid=2,
                        high=8,
                        confidence=0.98,
                    ),
                ),
                data_shape=DataShape(
                    estimated_state_size_gib=Interval(
                        low=100, mid=1000, high=4000, confidence=0.98
                    ),
                    # Cassandra compresses with LZ4 by default
                    estimated_compression_ratio=Interval(
                        low=2, mid=3, high=5, confidence=0.98
                    ),
                    # We dynamically allocate the C* JVM memory in the plan
                    # but account for the Priam sidecar here
                    reserved_instance_app_mem_gib=4,
                ),
                buffers=buffers,
            )


nflx_cassandra_capacity_model = NflxCassandraCapacityModel()



from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple

from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import CapacityRegretParameters
from service_capacity_modeling.interface import certain_float
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import FixedInterval
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Platform
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionContext

__common_regrets__ = frozenset(("spend", "disk", "mem"))


def _disk_regret(  # noqa: C901
    regret_params: CapacityRegretParameters,
    optimal_plan: CapacityPlan,
    proposed_plan: CapacityPlan,
) -> float:
    # type -> disk
    optimal_disk = {}
    plan_disk = {}

    for zonal in optimal_plan.requirements.zonal:
        typ = zonal.requirement_type
        if typ not in optimal_disk:
            optimal_disk[typ] = 0.0
        optimal_disk[typ] += zonal.disk_gib.mid

    for regional in optimal_plan.requirements.regional:
        typ = regional.requirement_type
        if typ not in optimal_disk:
            optimal_disk[typ] = 0.0
        optimal_disk[typ] += regional.disk_gib.mid

    for zonal in proposed_plan.requirements.zonal:
        typ = zonal.requirement_type
        if typ not in plan_disk:
            plan_disk[typ] = 0.0
        plan_disk[typ] += zonal.disk_gib.mid

    for regional in proposed_plan.requirements.regional:
        typ = regional.requirement_type
        if typ not in plan_disk:
            plan_disk[typ] = 0.0
        plan_disk[typ] += regional.disk_gib.mid

    # We regret not having the disk space for a dataset, but do not
    # regret lacking disk space
    regret = 0.0
    for typ in optimal_disk.keys() | plan_disk.keys():
        od, pd = optimal_disk.get(typ, 0), plan_disk.get(typ, 0)
        if od > pd:
            regret += (
                (od - pd) * regret_params.disk.under_provision_cost
            ) ** regret_params.disk.exponent
    return regret


class CapacityModel:
    """Stateless interface for defining a capacity model

    To define a capacity model you must implement two pure functions.

    The first, `capacity_plan` method calculates the best possible
    CapacityPlan (cluster layout) given a concrete `Instance` shape (cpu,
    memory, disk etc ...) and attachable `Drive` (e.g. cloud drives like GP2)
    along with a concrete instead of CapacityDesires (qps, latency, etc ...).

    The second, `regret` method calculates a relative regret (higher is worse)
    of two capacity plans. For example if the optimal plan computed by
    `capacity_plan` would allocate `100` CPUs and the proposed plan only
    allocates `50` we regret that choice as we are under-provisioning.

    """

    def __init__(self):
        pass

    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:
        """Given a concrete hardware shape and desires, return a candidate

        This is the only required method on this interface. Your model
        must either:
            * Return None to indicate there is no viable use of the
              instance/drive which meets the user's desires
            * Return a CapapacityPlan containing the model's calculation
              of how much CPU/RAM/disk etc ... that is required and
        """
        # quiet pylint
        (_, _, _, _, _) = (instance, drive, context, desires, extra_model_arguments)
        return None

    @staticmethod
    def regret(
        regret_params: CapacityRegretParameters,
        optimal_plan: CapacityPlan,
        proposed_plan: CapacityPlan,
    ) -> Dict[str, float]:
        """Optional cost model for how much we regret a choice

        After the capacity planner has simulated a bunch of possible outcomes
        We need to evaluate each cluster against the optimal choice for a
        given requirement.

        Our default model of cost is just related to money and disk footprint:

        Under an assumption that cost is a reasonable single dimension
        to compare to clusters on, we penalize under-provisioned (cheap)
        clusters more than expensive ones.

        :return: A componentized regret function where the total regret is
        the sum of all componenets. This is not just a single number so as you
        develop more complex regret functions you can debug why clusters are
        or are not being chosen
        """
        regrets = {"spend": 0.0, "disk": 0.0, "mem": 0.0}

        # Have to subtract out service costs which should be proprotional
        # to the input not the output.
        optimal_service_spend = sum(
            s.annual_cost
            for s in optimal_plan.candidate_clusters.services
            if s.regret_cost is False
        )
        optimal_cost = (
            float(optimal_plan.candidate_clusters.total_annual_cost)
            - optimal_service_spend
        )
        plan_service_spend = sum(
            s.annual_cost
            for s in proposed_plan.candidate_clusters.services
            if s.regret_cost is False
        )
        plan_cost = (
            float(proposed_plan.candidate_clusters.total_annual_cost)
            - plan_service_spend
        )

        if "spend" in optimal_plan.requirements.regrets:
            if plan_cost >= optimal_cost:
                regrets["spend"] = (
                    (plan_cost - optimal_cost) * regret_params.spend.over_provision_cost
                ) ** regret_params.spend.exponent
            else:
                regrets["spend"] = (
                    (optimal_cost - plan_cost)
                    * regret_params.spend.under_provision_cost
                ) ** regret_params.spend.exponent

        if "disk" in optimal_plan.requirements.regrets:
            regrets["disk"] = _disk_regret(
                regret_params=regret_params,
                optimal_plan=optimal_plan,
                proposed_plan=proposed_plan,
            )

        if "mem" in optimal_plan.requirements.regrets:
            optimal_mem = sum(
                req.mem_gib.mid for req in optimal_plan.requirements.zonal
            )
            optimal_mem += sum(
                req.mem_gib.mid for req in optimal_plan.requirements.regional
            )

            plan_mem = sum(req.mem_gib.mid for req in proposed_plan.requirements.zonal)
            plan_mem += sum(
                req.mem_gib.mid for req in proposed_plan.requirements.regional
            )

            # Running out of memory is particularly costly because it often
            # can cause an outage that is hard to get out of. We do not regret
            # too much memory
            if optimal_mem > plan_mem:
                regrets["mem"] = (
                    (optimal_mem - plan_mem) * regret_params.mem.under_provision_cost
                ) ** regret_params.mem.exponent

        for regret in optimal_plan.requirements.regrets:
            if regret not in __common_regrets__:
                regrets["regret"] = optimal_plan.requirements.regret(
                    name=regret, optimal_plan=optimal_plan, proposed_plan=proposed_plan
                )

        return regrets

    @staticmethod
    def description() -> str:
        """Optional description of the model"""
        return "No description"

    @staticmethod
    def extra_model_arguments_schema() -> Dict[str, Any]:
        """Optional json schema of extra keyword arguments

        Some models might take additional arguments to capacity_plan.
        They can convey that context to callers here along with a
        description of each argument
        """
        return {"type": "object"}

    @staticmethod
    def run_hardware_simulation() -> bool:
        """Optional to skip hardware simulation

        Some models, managed services, do not
        require simulating through hardware (instances, drives).

        """
        return True

    @staticmethod
    def compose_with(
        user_desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Tuple[Tuple[str, Callable[[CapacityDesires], CapacityDesires]], ...]:
        """Return additional model names to compose with this one

        The second element of the tuple is a capacity desire transform that
        takes the original user desire and modifies it for the composed
        model.

        (("model1", lambda x: x),
         ("model2", lambda x: transform(x)))

        Often used for dependencies.
        """
        # quiet pylint
        (_, _) = user_desires, extra_model_arguments
        return tuple()

    @staticmethod
    def allowed_platforms() -> Tuple[Platform, ...]:
        """Return which platforms this model accepts.

        Most software can run on amd64 (Intel and AMD), but some models might
        accept others
        """
        return (Platform.amd64,)

    @staticmethod
    def allowed_cloud_drives() -> Tuple[Optional[str], ...]:
        """Return which cloud drives this model accepts.

        This will _override_ the lifecycle of the drive.

        Note that an empty tuple means _all_ drives. If you want to
        only accept local hard drives emit None
        """
        return tuple()

    @staticmethod
    def default_desires(
        user_desires: CapacityDesires, extra_model_arguments: Dict[str, Any]
    ):
        """Optional defaults to apply given a user desires

        Often users do not know what the on-cpu time of their queries
        will be, but the models often have a good idea. For example
        databases usually have some range of on-cpu time for point queries
        (latency access) versus throughput.

        This is also a good place to throw ValueError on AccessPattern
        or AccessConsistency that cannot be met.
        """
        _ = extra_model_arguments

        unlikely_consistency_models = (
            AccessConsistency.linearizable,
            AccessConsistency.serializable,
        )
        target = (
            user_desires.query_pattern.access_consistency.same_region.target_consistency
        )
        if target in unlikely_consistency_models:
            raise ValueError(
                f"Most services do not support {unlikely_consistency_models}"
            )

        if user_desires.query_pattern.access_pattern == AccessPattern.latency:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.latency,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.read_your_writes,
                            staleness_slo_sec=FixedInterval(low=0, mid=0.1, high=1),
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.best_effort,
                            staleness_slo_sec=FixedInterval(low=10, mid=60, high=600),
                        ),
                    ),
                    estimated_mean_read_latency_ms=certain_float(1),
                    estimated_mean_write_latency_ms=certain_float(1),
                    # "Single digit milliseconds"
                    read_latency_slo_ms=FixedInterval(
                        low=0.4, mid=4, high=10, confidence=0.98
                    ),
                    write_latency_slo_ms=FixedInterval(
                        low=0.4, mid=4, high=10, confidence=0.98
                    ),
                ),
                data_shape=DataShape(),
            )
        else:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.throughput,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.read_your_writes,
                            staleness_slo_sec=FixedInterval(low=0, mid=0.1, high=1),
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.best_effort,
                            staleness_slo_sec=FixedInterval(low=10, mid=60, high=600),
                        ),
                    ),
                    estimated_mean_read_latency_ms=certain_float(2),
                    estimated_mean_write_latency_ms=certain_float(4),
                    # "Tens of milliseconds"
                    read_latency_slo_ms=FixedInterval(
                        low=10, mid=50, high=100, confidence=0.98
                    ),
                    write_latency_slo_ms=FixedInterval(
                        low=10, mid=50, high=100, confidence=0.98
                    ),
                ),
                data_shape=DataShape(),
            )



from functools import lru_cache
from typing import Sequence
from typing import Tuple

import numpy as np
from scipy.optimize import minimize
from scipy.optimize import OptimizeResult
from scipy.optimize import root
from scipy.special import gammainc as gammaf
from scipy.stats import beta as beta_dist
from scipy.stats import gamma as gamma_dist
from scipy.stats import rv_continuous

from service_capacity_modeling.interface import certain_float
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import IntervalModel

# Parameter estimation of various scipy distributions using
# See https://www.johndcook.com/quantiles_parameters.pdf for
# background

EPSILON = 0.001

# Gamma distribution G(alpha, beta) with mean alpha * beta


def _gamma_fn_from_params(low, mid, high, confidence):
    assert 0 < low <= mid <= high
    confidence = min(confidence, 0.99)
    confidence = max(confidence, 0.01)

    low_p = 0 + (1 - confidence) / 2.0
    high_p = 1 - (1 - confidence) / 2.0

    # cdf(x) = F(k) * gammaf(shape, x / scale)
    # mean = shape * scale
    # We know the value at two points of the cdf and the mean so we can
    # basically setup a system of equations of cdf(high) / cdf(low) = known
    # and mean = known
    #
    # Then we can use numeric methods to solve for the remaining shape parameter

    def f(k):
        zero = high / low
        return gammaf(k, high_p * k / mid) / gammaf(k, low_p * k / mid) - zero

    return f


def _gamma_dist_from_interval(
    interval: Interval, seed: int = 0xCAFE
) -> Tuple[float, rv_continuous]:
    # If we know cdf(high), cdf(low) and mean (mid) we can use an iterative
    # solver to find a possible gamma interval

    # Note we shift the lower bound and mean by the minimum (defaults to
    # half the lower bound so we don't end up with less than the minimum
    # estimate. This does distort the gamma but in a way that is useful for
    # capacity planning (sorta like a wet-bias in forcasting models)
    minimum = interval.minimum
    lower = interval.low - minimum
    mean = interval.mid - minimum

    if lower == 0:
        lower = EPSILON

    f = _gamma_fn_from_params(lower, mean, interval.high, interval.confidence)
    result = root(f, 2)
    shape = result.x[0]

    dist = gamma_dist(shape, loc=minimum, scale=mean / shape)
    dist.random_state = np.random.default_rng(seed=seed)
    return (shape, dist)


# This can be expensive, so cache it
@lru_cache(maxsize=128)
def _gamma_for_interval(interval: Interval, seed: int = 0xCAFE) -> rv_continuous:
    return _gamma_dist_from_interval(interval, seed=seed)[1]


def gamma_for_interval(interval: Interval, seed: int = 0xCAFE) -> rv_continuous:
    result = _gamma_for_interval(interval, seed)
    # Use the new Generator API instead of RandomState for ~20% speedup
    result.random_state = np.random.default_rng(seed=seed)
    return result


# Beta distribution B(alpha, beta) with mean alpha / (alpha + beta)


def _beta_cost_fn_from_params(low, mid, high, confidence):
    assert low <= mid <= high < 1.0
    assert mid > 0

    # Assume symmetric percentiles were provided
    confidence = min(confidence, 0.99)
    confidence = max(confidence, 0.01)

    low_p = 0.0 + (1 - confidence) / 2.0
    high_p = 1.0 - (1 - confidence) / 2.0

    def cost(alpha):
        beta = alpha / mid - alpha
        if alpha == 0 or beta == 0:
            return float("inf")

        cost = (beta_dist.cdf(low, alpha, beta) - low_p) ** 2
        cost += (beta_dist.cdf(high, alpha, beta) - high_p) ** 2
        return cost

    return cost


def _beta_dist_from_interval(
    interval: Interval, seed: int = 0xCAFE
) -> Tuple[Tuple[float, float, OptimizeResult], rv_continuous]:
    # If we know cdf(high), cdf(low) and mean (mid) we can use an iterative
    # solver to find a possible beta fit

    if interval.minimum == interval.maximum:
        minimum = interval.low - EPSILON
        maximum = interval.high + EPSILON
        scale = maximum - minimum
    else:
        minimum = interval.minimum
        maximum = interval.maximum
        scale = maximum - minimum

    lower = (interval.low - minimum) / scale
    mean = (interval.mid - minimum) / scale
    upper = (interval.high - minimum) / scale

    f = _beta_cost_fn_from_params(lower, mean, upper, interval.confidence)
    result = minimize(f, x0=2, bounds=[(0.1, 40)])
    alpha = result.x[0]

    dist = beta_dist(alpha, alpha / mean - alpha, loc=minimum, scale=scale)
    dist.random_state = np.random.default_rng(seed=seed)
    return (alpha, alpha / mean - alpha, result), dist


# This can be expensive, so cache it
@lru_cache(maxsize=128)
def _beta_for_interval(interval: Interval, seed: int = 0xCAFE) -> rv_continuous:
    return _beta_dist_from_interval(interval, seed=seed)[1]


def beta_for_interval(interval: Interval, seed: int = 0xCAFE) -> rv_continuous:
    result = _beta_for_interval(interval, seed)
    # Use the new Generator API instead of RandomState for ~20% speedup
    result.random_state = np.random.default_rng(seed=seed)
    return result


def dist_for_interval(interval: Interval, seed: int = 0xCAFE) -> rv_continuous:
    if interval.model_with == IntervalModel.beta:
        result = beta_for_interval(interval=interval, seed=seed)
    elif interval.model_with == IntervalModel.gamma:
        result = gamma_for_interval(interval=interval, seed=seed)
    else:
        result = beta_for_interval(interval=interval, seed=seed)
    return result


def interval_percentile(
    interval: Interval, percentiles: Sequence[int]
) -> Sequence[Interval]:
    if interval.can_simulate:
        samples = dist_for_interval(interval).rvs(1028)
        p = np.percentile(samples, percentiles)
        return [certain_float(i) for i in p]
    else:
        return [interval] * len(percentiles)
